from __future__ import annotations

import hashlib
import logging
from typing import List, Union

from great_expectations.exceptions import exceptions as gx_exceptions
from great_expectations.execution_engine.split_and_sample.data_splitter import (
    DataSplitter,
    DatePart,
)

logger = logging.getLogger(__name__)

try:
    import pyspark
    import pyspark.sql.functions as F

    # noinspection SpellCheckingInspection
    import pyspark.sql.types as sparktypes
    from pyspark.sql import DataFrame
except ImportError:
    pyspark = None
    F = None
    DataFrame = None
    # noinspection SpellCheckingInspection
    sparktypes = None
    logger.debug(
        "Unable to load pyspark; install optional spark dependency if you will be working with Spark dataframes"
    )


class SparkDataSplitter(DataSplitter):
    """Methods for splitting data accessible via SparkDFExecutionEngine.

    Note, for convenience, you can also access DatePart via the instance variable
    date_part e.g. SparkDataSplitter.date_part.MONTH
    """

    def split_on_year(
        self,
        df: DataFrame,
        column_name: str,
        batch_identifiers: dict,
    ) -> DataFrame:
        """Split on year values in column_name.

        Args:
            df: dataframe from batch data.
            column_name: column in table to use in determining split.
            batch_identifiers: should contain a dateutil parseable datetime whose
                relevant date parts will be used for splitting or key values
                of {date_part: date_part_value}.

        Returns:
            List of boolean clauses based on whether the date_part value in the
                batch identifier matches the date_part value in the column_name column.
        """
        return self.split_on_date_parts(
            df=df,
            column_name=column_name,
            batch_identifiers=batch_identifiers,
            date_parts=[DatePart.YEAR],
        )

    def split_on_year_and_month(
        self,
        df: DataFrame,
        column_name: str,
        batch_identifiers: dict,
    ) -> DataFrame:
        """Split on year and month values in column_name.

        Args:
            df: dataframe from batch data.
            column_name: column in table to use in determining split.
            batch_identifiers: should contain a dateutil parseable datetime whose
                relevant date parts will be used for splitting or key values
                of {date_part: date_part_value}.

        Returns:
            List of boolean clauses based on whether the date_part value in the
                batch identifier matches the date_part value in the column_name column.
        """
        return self.split_on_date_parts(
            df=df,
            column_name=column_name,
            batch_identifiers=batch_identifiers,
            date_parts=[DatePart.YEAR, DatePart.MONTH],
        )

    def split_on_year_and_month_and_day(
        self,
        df: DataFrame,
        column_name: str,
        batch_identifiers: dict,
    ) -> DataFrame:
        """Split on year and month and day values in column_name.

        Args:
            df: dataframe from batch data.
            column_name: column in table to use in determining split.
            batch_identifiers: should contain a dateutil parseable datetime whose
                relevant date parts will be used for splitting or key values
                of {date_part: date_part_value}.

        Returns:
            List of boolean clauses based on whether the date_part value in the
                batch identifier matches the date_part value in the column_name column.
        """
        return self.split_on_date_parts(
            df=df,
            column_name=column_name,
            batch_identifiers=batch_identifiers,
            date_parts=[DatePart.YEAR, DatePart.MONTH, DatePart.DAY],
        )

    def split_on_date_parts(
        self,
        df: DataFrame,
        column_name: str,
        batch_identifiers: dict,
        date_parts: Union[List[DatePart], List[str]],
    ) -> DataFrame:
        """Split on date_part values in column_name.

        Values are NOT truncated, for example this will return data for a
        given month (if only month is chosen for date_parts) for ALL years.
        This may be useful for viewing seasonality, but you can also specify
        multiple date_parts to achieve date_trunc like behavior e.g.
        year, month and day.

        Args:
            df: dataframe from batch data.
            column_name: column in data used to determine split.
            batch_identifiers: should contain a dateutil parseable datetime whose date parts
                will be used for splitting or key values of {date_part: date_part_value}
            date_parts: part of the date to be used for splitting e.g.
                DatePart.DAY or the case-insensitive string representation "day"

        Returns:
            Dataframe with splitting applied.
        """
        self._validate_date_parts(date_parts)

        date_parts = self._convert_date_parts(date_parts)

        column_batch_identifiers: dict = batch_identifiers[column_name]

        date_parts_dict: dict = (
            self._convert_datetime_batch_identifiers_to_date_parts_dict(
                column_batch_identifiers, date_parts
            )
        )

        for date_part, date_part_value in date_parts_dict.items():
            df = df.filter(
                getattr(F, self._convert_date_part_to_spark_equivalent(date_part))(
                    F.col(column_name)
                )
                == date_part_value
            )
        return df

    @staticmethod
    def _convert_date_part_to_spark_equivalent(date_part: DatePart | str) -> str:
        """Convert the DatePart to a string representing the corresponding pyspark.sql.functions version.

        For example DatePart.DAY -> pyspark.sql.functions.dayofmonth() -> "dayofmonth"

        Args:
            date_part: DatePart representing the part of the datetime to extract or string equivalent.

        Returns:
            String representing the spark function to use for the given DatePart.
        """
        date_part = DatePart(date_part)

        spark_date_part_decoder: dict = {
            DatePart.YEAR: "year",
            DatePart.MONTH: "month",
            DatePart.WEEK: "weekofyear",
            DatePart.DAY: "dayofmonth",
            DatePart.HOUR: "hour",
            DatePart.MINUTE: "minute",
            DatePart.SECOND: "second",
        }
        return spark_date_part_decoder[date_part]

    @staticmethod
    def split_on_whole_table(
        df: DataFrame,
    ) -> DataFrame:
        """No op. Return the same data that is passed in.

        Args:
            df: Spark DataFrame that will be returned

        Returns:
            Unfiltered DataFrame.
        """
        return df

    @staticmethod
    def split_on_column_value(
        df, column_name: str, batch_identifiers: dict
    ) -> DataFrame:
        """Return a dataframe where rows are filtered based on the specified column value.

        Args:
            df: Spark DataFrame to be filtered.
            column_name: Column to use in comparison.
            batch_identifiers: Contains value to use in comparison e.g. batch_identifiers={ 'col': value }.

        Returns:
            Filtered spark DataFrame.
        """
        return df.filter(F.col(column_name) == batch_identifiers[column_name])

    @staticmethod
    def split_on_converted_datetime(
        df,
        column_name: str,
        batch_identifiers: dict,
        date_format_string: str = "yyyy-MM-dd",
    ) -> DataFrame:
        """Return a dataframe where rows are filtered based on whether their converted
        datetime (using date_format_string) matches the datetime string value provided
        in batch_identifiers for the specified column.

        Args:
            df: Spark DataFrame to be filtered.
            column_name: Column to use in comparison.
            batch_identifiers: Value to use in comparison as {column_name: datetime string}.
            date_format_string: Format used to convert datetime column for comparison to
                batch identifiers.

        Returns:
            Filtered spark DataFrame.
        """
        matching_string = batch_identifiers[column_name]
        res = (
            df.withColumn(
                "date_time_tmp", F.from_unixtime(F.col(column_name), date_format_string)
            )
            .filter(F.col("date_time_tmp") == matching_string)
            .drop("date_time_tmp")
        )
        return res

    @staticmethod
    def split_on_divided_integer(
        df, column_name: str, divisor: int, batch_identifiers: dict
    ):
        """Divide the values in the named column by `divisor`, and split on that"""
        matching_divisor = batch_identifiers[column_name]
        res = (
            df.withColumn(
                "div_temp",
                (F.col(column_name) / divisor).cast(sparktypes.IntegerType()),
            )
            .filter(F.col("div_temp") == matching_divisor)
            .drop("div_temp")
        )
        return res

    @staticmethod
    def split_on_mod_integer(df, column_name: str, mod: int, batch_identifiers: dict):
        """Divide the values in the named column by `divisor`, and split on that"""
        matching_mod_value = batch_identifiers[column_name]
        res = (
            df.withColumn(
                "mod_temp", (F.col(column_name) % mod).cast(sparktypes.IntegerType())
            )
            .filter(F.col("mod_temp") == matching_mod_value)
            .drop("mod_temp")
        )
        return res

    @staticmethod
    def split_on_multi_column_values(df, column_names: list, batch_identifiers: dict):
        """Split on the joint values in the named columns"""
        for column_name in column_names:
            value = batch_identifiers.get(column_name)
            if not value:
                raise ValueError(
                    f"In order for SparkDFExecutionEngine to `_split_on_multi_column_values`, "
                    f"all values in  column_names must also exist in batch_identifiers. "
                    f"{column_name} was not found in batch_identifiers."
                )
            df = df.filter(F.col(column_name) == value)
        return df

    @staticmethod
    def split_on_hashed_column(
        df,
        column_name: str,
        hash_digits: int,
        batch_identifiers: dict,
        hash_function_name: str = "sha256",
    ):
        """Split on the hashed value of the named column"""
        try:
            getattr(hashlib, hash_function_name)
        except (TypeError, AttributeError):
            raise (
                gx_exceptions.ExecutionEngineError(
                    f"""The splitting method used with SparkDFExecutionEngine has a reference to an invalid hash_function_name.
                    Reference to {hash_function_name} cannot be found."""
                )
            )

        def _encrypt_value(to_encode):
            hash_func = getattr(hashlib, hash_function_name)
            hashed_value = hash_func(to_encode.encode()).hexdigest()[-1 * hash_digits :]
            return hashed_value

        encrypt_udf = F.udf(_encrypt_value, sparktypes.StringType())
        res = (
            df.withColumn("encrypted_value", encrypt_udf(column_name))
            .filter(F.col("encrypted_value") == batch_identifiers["hash_value"])
            .drop("encrypted_value")
        )
        return res



# Test File Expectations
import os

import pytest

import great_expectations as gx
from great_expectations.data_context.util import file_relative_path


def test_expect_file_line_regex_match_count_to_be_between():
    #####Invalid File Path######
    joke_file_path = "joke.txt"
    assert not os.path.isfile(joke_file_path)
    joke_dat = gx.data_asset.FileDataAsset(joke_file_path)

    with pytest.raises(IOError):
        joke_dat.expect_file_line_regex_match_count_to_be_between(
            regex=r",\S", expected_min_count=0, expected_max_count=4, skip=1
        )

    complete_file_path = file_relative_path(
        __file__, "../test_sets/toy_data_complete.csv"
    )
    file_dat = gx.data_asset.FileDataAsset(complete_file_path)

    # Invalid Skip Parameter
    with pytest.raises(ValueError):
        file_dat.expect_file_line_regex_match_count_to_be_between(
            regex=r",\S", expected_min_count=0, expected_max_count=4, skip=2.4
        )

    # Invalid Regex
    with pytest.raises(ValueError):
        file_dat.expect_file_line_regex_match_count_to_be_between(
            regex=2, expected_min_count=1, expected_max_count=8, skip=2
        )

    # Non-integer min value
    with pytest.raises(ValueError):
        file_dat.expect_file_line_regex_match_count_to_be_between(
            regex=r",\S", expected_min_count=1.3, expected_max_count=8, skip=1
        )

    # Negative min value
    with pytest.raises(ValueError):
        file_dat.expect_file_line_regex_match_count_to_be_between(
            regex=r",\S", expected_min_count=-2, expected_max_count=8, skip=1
        )

    # Non-integer max value
    with pytest.raises(ValueError):
        file_dat.expect_file_line_regex_match_count_to_be_between(
            regex=r",\S", expected_min_count=0, expected_max_count="foo", skip=1
        )

    # Negative max value
    with pytest.raises(ValueError):
        file_dat.expect_file_line_regex_match_count_to_be_between(
            regex=r",\S", expected_min_count=0, expected_max_count=-1, skip=1
        )

    # Min count more than max count
    with pytest.raises(ValueError):
        file_dat.expect_file_line_regex_match_count_to_be_between(
            regex=r",\S", expected_min_count=4, expected_max_count=3, skip=1
        )

    # Count does not fall in range
    fail_trial = file_dat.expect_file_line_regex_match_count_to_be_between(
        regex=r",\S", expected_min_count=9, expected_max_count=12, skip=1
    )

    assert not fail_trial.success
    assert fail_trial.result["unexpected_percent"] == 100
    assert fail_trial.result["missing_percent"] == 0

    # Count does fall in range
    success_trial = file_dat.expect_file_line_regex_match_count_to_be_between(
        regex=r",\S", expected_min_count=0, expected_max_count=4, skip=1
    )

    assert success_trial.success
    assert success_trial.result["unexpected_percent"] == 0
    assert success_trial.result["missing_percent"] == 0


def test_expect_file_line_regex_match_count_to_equal():
    complete_file_path = file_relative_path(
        __file__, "../test_sets/toy_data_complete.csv"
    )
    incomplete_file_path = file_relative_path(
        __file__, "../test_sets/toy_data_incomplete.csv"
    )
    file_dat = gx.data_asset.FileDataAsset(complete_file_path)
    file_incomplete_dat = gx.data_asset.FileDataAsset(incomplete_file_path)

    # Invalid Regex Value
    with pytest.raises(ValueError):
        file_dat.expect_file_line_regex_match_count_to_equal(
            regex=True, expected_count=3, skip=1
        )

    # Non-integer expected_count
    with pytest.raises(ValueError):
        file_dat.expect_file_line_regex_match_count_to_equal(
            regex=r",\S", expected_count=6.3, skip=1
        )

    # Negative expected_count
    with pytest.raises(ValueError):
        file_dat.expect_file_line_regex_match_count_to_equal(
            regex=r",\S", expected_count=-6, skip=1
        )

    # Count does not equal expected count
    fail_trial = file_incomplete_dat.expect_file_line_regex_match_count_to_equal(
        regex=r",\S", expected_count=3, skip=1
    )

    assert not fail_trial.success
    assert fail_trial.result["unexpected_percent"] == (3 / 7 * 100)
    assert fail_trial.result["unexpected_percent_total"] == (3 / 9 * 100)
    assert fail_trial.result["missing_percent"] == (2 / 9 * 100)
    assert fail_trial.result["unexpected_percent_nonmissing"] == (3 / 7 * 100)

    # Mostly success
    mostly_trial = file_incomplete_dat.expect_file_line_regex_match_count_to_equal(
        regex=r",\S", expected_count=3, skip=1, mostly=0.57
    )

    assert mostly_trial.success

    # Count does fall in range
    success_trial = file_dat.expect_file_line_regex_match_count_to_equal(
        regex=r",\S", expected_count=3, skip=1
    )

    assert success_trial.success
    assert success_trial.result["unexpected_percent"] == 0
    assert success_trial.result["unexpected_percent_nonmissing"] == 0
    assert success_trial.result["missing_percent"] == 0


def test_expect_file_hash_to_equal():
    # Test for non-existent file
    fake_file = gx.data_asset.FileDataAsset(file_path="abc")

    with pytest.raises(IOError):
        fake_file.expect_file_hash_to_equal(value="abc")

    # Test for non-existent hash algorithm
    titanic_path = file_relative_path(__file__, "../test_sets/Titanic.csv")
    titanic_file = gx.data_asset.FileDataAsset(titanic_path)

    with pytest.raises(ValueError):
        titanic_file.expect_file_hash_to_equal(hash_alg="md51", value="abc")

    # Test non-matching hash value
    fake_hash_value = titanic_file.expect_file_hash_to_equal(value="abc")
    assert not fake_hash_value.success

    # Test matching hash value with default algorithm
    hash_value = "63188432302f3a6e8c9e9c500ff27c8a"
    good_hash_default_alg = titanic_file.expect_file_hash_to_equal(value=hash_value)
    assert good_hash_default_alg.success

    # Test matching hash value with specified algorithm
    hash_value = "f89f46423b017a1fc6a4059d81bddb3ff64891e3c81250fafad6f3b3113ecc9b"
    good_hash_new_alg = titanic_file.expect_file_hash_to_equal(
        value=hash_value, hash_alg="sha256"
    )
    assert good_hash_new_alg.success


def test_expect_file_size_to_be_between():
    fake_file = gx.data_asset.FileDataAsset("abc")

    # Test for non-existent file
    with pytest.raises(OSError):
        fake_file.expect_file_size_to_be_between(0, 10000)

    titanic_path = file_relative_path(__file__, "../test_sets/Titanic.csv")
    titanic_file = gx.data_asset.FileDataAsset(titanic_path)

    # Test minsize not an integer
    with pytest.raises(ValueError):
        titanic_file.expect_file_size_to_be_between("a", 10000)

    # Test maxsize not an integer
    with pytest.raises(ValueError):
        titanic_file.expect_file_size_to_be_between(0, "10000a")

    # Test minsize less than 0
    with pytest.raises(ValueError):
        titanic_file.expect_file_size_to_be_between(-1, 10000)

    # Test maxsize less than 0
    with pytest.raises(ValueError):
        titanic_file.expect_file_size_to_be_between(0, -1)

    # Test minsize > maxsize
    with pytest.raises(ValueError):
        titanic_file.expect_file_size_to_be_between(10000, 0)

    # Test file size not in range
    bad_range = titanic_file.expect_file_size_to_be_between(0, 10000)
    assert not bad_range.success

    # Test file size in range
    lower, upper = (70000, 72000)
    good_range = titanic_file.expect_file_size_to_be_between(lower, upper)
    assert good_range.success


def test_expect_file_to_exist():
    # Test for non-existent file
    fake_file = gx.data_asset.FileDataAsset("abc")
    fake_file_existence = fake_file.expect_file_to_exist()
    assert not fake_file_existence.success

    # Test for existing file
    real_file = gx.data_asset.FileDataAsset(
        file_relative_path(__file__, "../test_sets/Titanic.csv")
    )
    real_file_existence = real_file.expect_file_to_exist()
    assert real_file_existence.success


def test_expect_file_to_have_valid_table_header():
    # Test for non-existent file
    fake_file = gx.data_asset.FileDataAsset("abc")
    with pytest.raises(IOError):
        fake_file.expect_file_to_have_valid_table_header(regex="")

    # Test for non-unique column names
    invalid_header_dat = gx.data_asset.FileDataAsset(
        file_relative_path(__file__, "../test_sets/same_column_names.csv")
    )
    invalid_header_dat_expectation = (
        invalid_header_dat.expect_file_to_have_valid_table_header(regex=r"\|", skip=2)
    )
    assert not invalid_header_dat_expectation.success

    # Test for unique column names
    valid_header_dat = gx.data_asset.FileDataAsset(
        file_relative_path(__file__, "../test_sets/Titanic.csv")
    )
    valid_header_dat_expectation = (
        valid_header_dat.expect_file_to_have_valid_table_header(regex=",")
    )
    assert valid_header_dat_expectation.success


def test_expect_file_to_be_valid_json():
    # Test for non-existent file
    fake_file = gx.data_asset.FileDataAsset("abc")
    with pytest.raises(IOError):
        fake_file.expect_file_to_be_valid_json()

    # Test invalid JSON file
    invalid_JSON_file = gx.data_asset.FileDataAsset(
        file_relative_path(__file__, "../test_sets/invalid_json_file.json")
    )
    invalid_JSON_expectation = invalid_JSON_file.expect_file_to_be_valid_json()
    assert not invalid_JSON_expectation.success

    # Test valid JSON file
    valid_JSON_file = gx.data_asset.FileDataAsset(
        file_relative_path(__file__, "../test_sets/titanic_expectations.json")
    )
    valid_JSON_expectation = valid_JSON_file.expect_file_to_be_valid_json()
    assert valid_JSON_expectation.success

    # Test valid JSON file with non-matching schema
    schema_file = file_relative_path(__file__, "../test_sets/sample_schema.json")
    test_file = gx.data_asset.FileDataAsset(
        file_relative_path(__file__, "../test_sets/json_test1_against_schema.json")
    )
    test_file_expectation = test_file.expect_file_to_be_valid_json(schema=schema_file)
    assert not test_file_expectation.success

    # Test valid JSON file with valid schema
    test_file = gx.data_asset.FileDataAsset(
        file_relative_path(__file__, "../test_sets/json_test2_against_schema.json")
    )
    schema_file = file_relative_path(__file__, "../test_sets/sample_schema.json")
    test_file_expectation = test_file.expect_file_to_be_valid_json(schema=schema_file)
    assert test_file_expectation.success



import importlib
import itertools
import json
from collections.abc import Iterable
from typing import Any, Dict, List, Optional, Union

from marshmallow import ValidationError


class GreatExpectationsError(Exception):
    def __init__(self, message) -> None:
        self.message = message
        super().__init__(message)


class GreatExpectationsValidationError(ValidationError, GreatExpectationsError):
    def __init__(self, message, validation_error=None) -> None:
        self.message = message
        self.messages: Union[List[str], List[Any], Dict] = []
        if validation_error is not None:
            self.messages = validation_error.messages

    def __str__(self) -> str:
        if self.message is None:
            return str(self.messages)
        return self.message


class SuiteEditNotebookCustomTemplateModuleNotFoundError(ModuleNotFoundError):
    def __init__(self, custom_module) -> None:
        message = f"The custom module '{custom_module}' could not be found"
        super().__init__(message)


class DataContextError(GreatExpectationsError):
    pass


class ExpectationSuiteError(DataContextError):
    pass


class CheckpointError(DataContextError):
    pass


class CheckpointNotFoundError(CheckpointError):
    pass


class StoreBackendError(DataContextError):
    pass


class ParserError(GreatExpectationsError):
    pass


class InvalidConfigurationYamlError(DataContextError):
    pass


class InvalidTopLevelConfigKeyError(GreatExpectationsError):
    pass


class MissingTopLevelConfigKeyError(GreatExpectationsValidationError):
    pass


class RenderingError(GreatExpectationsError):
    pass


class InvalidBaseYamlConfigError(GreatExpectationsValidationError):
    def __init__(self, message, validation_error=None, field_name=None) -> None:
        if validation_error is not None:
            if (
                validation_error
                and validation_error.messages
                and isinstance(validation_error.messages, dict)
                and all([key is None for key in validation_error.messages.keys()])
            ):
                validation_error.messages = list(
                    itertools.chain.from_iterable(validation_error.messages.values())
                )
        super().__init__(message=message, validation_error=validation_error)
        self.field_name = field_name


class InvalidDataContextConfigError(InvalidBaseYamlConfigError):
    pass


class InvalidCheckpointConfigError(InvalidBaseYamlConfigError):
    pass


class InvalidBatchKwargsError(GreatExpectationsError):
    pass


class InvalidBatchSpecError(GreatExpectationsError):
    pass


class InvalidBatchRequestError(GreatExpectationsError):
    pass


class InvalidBatchIdError(GreatExpectationsError):
    pass


class InvalidDataContextKeyError(DataContextError):
    pass


class UnsupportedConfigVersionError(DataContextError):
    pass


class EvaluationParameterError(GreatExpectationsError):
    pass


class ProfilerError(GreatExpectationsError):
    pass


class ProfilerConfigurationError(ProfilerError):
    """A configuration error for a "RuleBasedProfiler" class."""

    pass


class ProfilerExecutionError(ProfilerError):
    """A runtime error for a "RuleBasedProfiler" class."""

    pass


class ProfilerNotFoundError(ProfilerError):
    pass


class DataAssistantError(ProfilerError):
    pass


class DataAssistantExecutionError(DataAssistantError):
    """A runtime error for a "DataAssistant" class."""

    pass


class DataAssistantResultExecutionError(DataAssistantError):
    """A runtime error for a "DataAssistantResult" class."""

    pass


class InvalidConfigError(DataContextError):
    def __init__(self, message) -> None:
        self.message = message
        super().__init__(self.message)


class MissingConfigVariableError(InvalidConfigError):
    def __init__(self, message, missing_config_variable=None) -> None:
        if not missing_config_variable:
            missing_config_variable = []
        self.message = message
        self.missing_config_variable = missing_config_variable
        super().__init__(self.message)


class AmbiguousDataAssetNameError(DataContextError):
    def __init__(self, message, candidates=None) -> None:
        self.message = message
        self.candidates = candidates
        super().__init__(self.message)


class StoreConfigurationError(DataContextError):
    pass


class InvalidExpectationKwargsError(GreatExpectationsError):
    pass


class InvalidExpectationConfigurationError(GreatExpectationsError):
    pass


class ExpectationNotFoundError(GreatExpectationsError):
    pass


class InvalidValidationResultError(GreatExpectationsError):
    pass


class GreatExpectationsTypeError(TypeError):
    pass


class StoreError(DataContextError):
    pass


class InvalidKeyError(StoreError):
    pass


class InvalidCacheValueError(GreatExpectationsError):
    def __init__(self, result_dict) -> None:
        template = """\
Invalid result values were found when trying to instantiate an ExpectationValidationResult.
- Invalid result values are likely caused by inconsistent cache values.
- Great Expectations enables caching by default.
- Please ensure that caching behavior is consistent between the underlying Dataset (e.g. Spark) and Great Expectations.
Result: {}
"""
        self.message = template.format(json.dumps(result_dict, indent=2))
        super().__init__(self.message)


class ConfigNotFoundError(DataContextError):
    """The great_expectations dir could not be found."""

    def __init__(self) -> None:
        self.message = """Error: No great_expectations directory was found here!
    - Please check that you are in the correct directory or have specified the correct directory.
    - If you have never run Great Expectations in this project, please run `great_expectations init` to get started.
"""
        super().__init__(self.message)


class PluginModuleNotFoundError(GreatExpectationsError):
    """A module import failed."""

    def __init__(self, module_name) -> None:
        template = """\
No module named `{}` could be found in your plugins directory.
    - Please verify your plugins directory is configured correctly.
    - Please verify you have a module named `{}` in your plugins directory.
"""
        self.message = template.format(module_name, module_name)

        colored_template = f"<red>{template}</red>"
        module_snippet = f"</red><yellow>{module_name}</yellow><red>"
        self.cli_colored_message = colored_template.format(
            module_snippet, module_snippet
        )
        super().__init__(self.message)


class PluginClassNotFoundError(DataContextError, AttributeError):
    """A module import failed."""

    def __init__(self, module_name, class_name) -> None:
        class_name_changes = {
            "FixedLengthTupleFilesystemStoreBackend": "TupleFilesystemStoreBackend",
            "FixedLengthTupleS3StoreBackend": "TupleS3StoreBackend",
            "FixedLengthTupleGCSStoreBackend": "TupleGCSStoreBackend",
            "InMemoryEvaluationParameterStore": "EvaluationParameterStore",
            "DatabricksTableGenerator": "DatabricksTableBatchKwargsGenerator",
            "GlobReaderGenerator": "GlobReaderBatchKwargsGenerator",
            "SubdirReaderGenerator": "SubdirReaderBatchKwargsGenerator",
            "QueryGenerator": "QueryBatchKwargsGenerator",
            "TableGenerator": "TableBatchKwargsGenerator",
            "S3Generator": "S3GlobReaderBatchKwargsGenerator",
            "ExtractAndStoreEvaluationParamsAction": "StoreEvaluationParametersAction",
            "StoreAction": "StoreValidationResultAction",
            "PartitionDefinitionSubset": "IDDict",
            "PartitionRequest": "IDDict",
            "PartitionDefinition": "IDDict",
            "PartitionQuery": "BatchFilter",
        }

        if class_name_changes.get(class_name):
            template = """The module: `{}` does not contain the class: `{}`.
            The class name `{}` has changed to `{}`."""
            self.message = template.format(
                module_name, class_name, class_name, class_name_changes.get(class_name)
            )
        else:
            template = """The module: `{}` does not contain the class: `{}`.
        - Please verify that the class named `{}` exists."""
            self.message = template.format(module_name, class_name, class_name)

        colored_template = f"<red>{template}</red>"
        module_snippet = f"</red><yellow>{module_name}</yellow><red>"
        class_snippet = f"</red><yellow>{class_name}</yellow><red>"
        if class_name_changes.get(class_name):
            new_class_snippet = (
                f"</red><yellow>{class_name_changes.get(class_name)}</yellow><red>"
            )
            self.cli_colored_message = colored_template.format(
                module_snippet, class_snippet, class_snippet, new_class_snippet
            )
        else:
            self.cli_colored_message = colored_template.format(
                module_snippet,
                class_snippet,
                class_snippet,
            )
        super().__init__(self.message)


class ClassInstantiationError(GreatExpectationsError):
    def __init__(self, module_name, package_name, class_name) -> None:
        # noinspection PyUnresolvedReferences
        module_spec: Optional[
            importlib.machinery.ModuleSpec
        ] = importlib.util.find_spec(module_name, package=package_name)
        if not module_spec:
            if not package_name:
                package_name = ""
            self.message = f"""No module named "{package_name + module_name}" could be found in the repository.  \
Please make sure that the file, corresponding to this package and module, exists and that dynamic loading of code \
modules, templates, and assets is supported in your execution environment.  This error is unrecoverable.
            """
        else:
            self.message = f"""The module "{module_name}" exists; however, the system is unable to create an instance \
of the class "{class_name}", searched for inside this module.  Please make sure that the class named "{class_name}" is \
properly defined inside its intended module and declared correctly by the calling entity.  This error is unrecoverable.
            """
        super().__init__(self.message)


class ExpectationSuiteNotFoundError(GreatExpectationsError):
    def __init__(self, data_asset_name) -> None:
        self.data_asset_name = data_asset_name
        self.message = (
            f"No expectation suite found for data_asset_name {data_asset_name}"
        )
        super().__init__(self.message)


class BatchKwargsError(DataContextError):
    def __init__(self, message, batch_kwargs=None) -> None:
        self.message = message
        self.batch_kwargs = batch_kwargs
        super().__init__(self.message)


class BatchDefinitionError(DataContextError):
    def __init__(self, message) -> None:
        self.message = message
        super().__init__(self.message)


class BatchSpecError(DataContextError):
    def __init__(self, message) -> None:
        self.message = message
        super().__init__(self.message)


class DatasourceError(DataContextError):
    def __init__(self, datasource_name: str, message: str) -> None:
        self.message = (
            f"Cannot initialize datasource {datasource_name}, error: {message}"
        )
        super().__init__(self.message)


class DatasourceConfigurationError(DatasourceError):
    pass


class DatasourceInitializationError(DatasourceError):
    pass


class DatasourceKeyPairAuthBadPassphraseError(DatasourceInitializationError):
    pass


class DatasourceNotFoundError(DataContextError):
    pass


class InvalidConfigValueTypeError(DataContextError):
    pass


class DataConnectorError(DataContextError):
    def __init__(self, message) -> None:
        self.message = message
        super().__init__(self.message)


class ExecutionEngineError(DataContextError):
    def __init__(self, message) -> None:
        self.message = message
        super().__init__(self.message)


class BatchFilterError(DataContextError):
    def __init__(self, message) -> None:
        self.message = message
        super().__init__(self.message)


class SorterError(DataContextError):
    def __init__(self, message) -> None:
        self.message = message
        super().__init__(self.message)


class SamplerError(DataContextError):
    pass


class MetricError(GreatExpectationsError):
    pass


class UnavailableMetricError(MetricError):
    pass


class MetricProviderError(MetricError):
    pass


class MetricComputationError(MetricError):
    pass


class InvalidMetricAccessorDomainKwargsKeyError(MetricError):
    pass


class MetricResolutionError(MetricError):
    def __init__(self, message, failed_metrics) -> None:
        super().__init__(message)
        if not isinstance(failed_metrics, Iterable):
            failed_metrics = (failed_metrics,)
        self.failed_metrics = failed_metrics


class GXCloudError(GreatExpectationsError):
    """
    Generic error used to provide additional context around Cloud-specific issues.
    """


class GXCloudConfigurationError(GreatExpectationsError):
    """
    Error finding and verifying the required configuration values when preparing to connect to GX Cloud
    """


class DatabaseConnectionError(GreatExpectationsError):
    """Error connecting to a database including during an integration test."""


class MigrationError(GreatExpectationsError):
    """Error when using the migration tool."""



import copy
import pathlib
from typing import Callable, List, Optional, cast
from unittest import mock

import pytest

import great_expectations.exceptions as gx_exceptions
from great_expectations.core.data_context_key import DataContextVariableKey
from great_expectations.core.serializer import (
    AbstractConfigSerializer,
    DictConfigSerializer,
    JsonConfigSerializer,
)
from great_expectations.core.yaml_handler import YAMLHandler
from great_expectations.data_context.cloud_constants import GXCloudRESTResource
from great_expectations.data_context.data_context.data_context import DataContext
from great_expectations.data_context.data_context_variables import (
    DataContextVariableSchema,
)
from great_expectations.data_context.store.datasource_store import DatasourceStore
from great_expectations.data_context.store.gx_cloud_store_backend import (
    GXCloudStoreBackend,
)
from great_expectations.data_context.types.base import (
    DatasourceConfig,
    datasourceConfigSchema,
)
from great_expectations.datasource.datasource_serializer import (
    JsonDatasourceConfigSerializer,
    NamedDatasourceSerializer,
    YAMLReadyDictDatasourceConfigSerializer,
)
from tests.data_context.conftest import MockResponse

yaml = YAMLHandler()


@pytest.fixture
def fake_datasource_name() -> str:
    return "my_first_datasource"


@pytest.fixture
def empty_datasource_store(datasource_store_name: str) -> DatasourceStore:
    return DatasourceStore(
        store_name=datasource_store_name,
        serializer=DictConfigSerializer(schema=datasourceConfigSchema),
    )


@pytest.fixture
def datasource_store_with_single_datasource(
    fake_datasource_name,
    datasource_config: DatasourceConfig,
    empty_datasource_store: DatasourceStore,
) -> DatasourceStore:
    key = DataContextVariableKey(
        resource_name=fake_datasource_name,
    )
    empty_datasource_store.set(key=key, value=datasource_config)
    return empty_datasource_store


@pytest.mark.unit
def test_datasource_store_with_bad_key_raises_error(
    empty_datasource_store: DatasourceStore, datasource_config: DatasourceConfig
) -> None:
    store: DatasourceStore = empty_datasource_store

    error_msg: str = "key must be an instance of DataContextVariableKey"

    with pytest.raises(TypeError) as e:
        store.set(key="my_bad_key", value=datasource_config)  # type: ignore[arg-type]
    assert error_msg in str(e.value)

    with pytest.raises(TypeError) as e:
        store.get(key="my_bad_key")  # type: ignore[arg-type]
    assert error_msg in str(e.value)


def _assert_serialized_datasource_configs_are_equal(
    datasource_configs: List[DatasourceConfig],
    serializers: Optional[List[AbstractConfigSerializer]] = None,
) -> None:
    """Assert that the datasource configs are equal using the DictConfigSerializer

    Args:
        datasource_configs: List of datasource configs to check

    Returns:
        None

    Raises:
        AssertionError
    """
    if len(datasource_configs) <= 1:
        raise AssertionError("Must provide at least 2 datasource configs")

    if serializers is None:
        serializers = [DictConfigSerializer(schema=datasourceConfigSchema)] * (
            len(datasource_configs) + 1
        )
    else:
        if len(serializers) <= 1:
            raise AssertionError("Must provide at least 2 datasource serializers")
        if not len(datasource_configs) == len(serializers):
            raise AssertionError(
                "Must provide the same number of serializers as datasource configs"
            )

    for idx, config in enumerate(datasource_configs[:-1]):
        assert serializers[idx].serialize(config) == serializers[idx + 1].serialize(
            datasource_configs[idx + 1]
        )


@pytest.mark.unit
def test__assert_serialized_datasource_configs_are_equal(
    datasource_config: DatasourceConfig, datasource_config_with_names: DatasourceConfig
) -> None:
    """Verify test helper method."""

    # Input errors:
    with pytest.raises(AssertionError):
        _assert_serialized_datasource_configs_are_equal([])

    with pytest.raises(AssertionError):
        _assert_serialized_datasource_configs_are_equal([datasource_config])

    # Happy path
    _assert_serialized_datasource_configs_are_equal(
        [datasource_config, datasource_config]
    )
    _assert_serialized_datasource_configs_are_equal(
        [datasource_config, datasource_config, datasource_config]
    )

    # Unequal configs
    with pytest.raises(AssertionError):
        _assert_serialized_datasource_configs_are_equal(
            [datasource_config, datasource_config_with_names]
        )

    with pytest.raises(AssertionError):
        _assert_serialized_datasource_configs_are_equal(
            [datasource_config, datasource_config, datasource_config_with_names]
        )

    with pytest.raises(AssertionError):
        _assert_serialized_datasource_configs_are_equal(
            [
                datasource_config,
                datasource_config,
                datasource_config_with_names,
                datasource_config,
            ]
        )


@pytest.mark.integration
def test_datasource_store_retrieval(
    empty_datasource_store: DatasourceStore, datasource_config: DatasourceConfig
) -> None:
    store: DatasourceStore = empty_datasource_store

    key = DataContextVariableKey(
        resource_name="my_datasource",
    )
    store.set(key=key, value=datasource_config)
    res: DatasourceConfig = store.get(key=key)

    assert isinstance(res, DatasourceConfig)
    set_config_serializer = DictConfigSerializer(schema=datasourceConfigSchema)
    retrieved_config_serializer = YAMLReadyDictDatasourceConfigSerializer(
        schema=datasourceConfigSchema
    )
    _assert_serialized_datasource_configs_are_equal(
        [datasource_config, res], [set_config_serializer, retrieved_config_serializer]
    )


@pytest.mark.integration
def test_datasource_store_set_cloud_mode(
    datasource_config: DatasourceConfig,
    datasource_config_with_names_and_ids: DatasourceConfig,
    mocked_datasource_post_response: Callable[[], MockResponse],
    mocked_datasource_get_response: Callable[[], MockResponse],
    ge_cloud_base_url: str,
    ge_cloud_access_token: str,
    ge_cloud_organization_id: str,
) -> None:
    ge_cloud_store_backend_config: dict = {
        "class_name": GXCloudStoreBackend.__name__,
        "ge_cloud_base_url": ge_cloud_base_url,
        "ge_cloud_resource_type": GXCloudRESTResource.DATASOURCE,
        "ge_cloud_credentials": {
            "access_token": ge_cloud_access_token,
            "organization_id": ge_cloud_organization_id,
        },
        "suppress_store_backend_id": True,
    }

    store = DatasourceStore(
        store_name="my_cloud_datasource_store",
        store_backend=ge_cloud_store_backend_config,
        serializer=JsonConfigSerializer(schema=datasourceConfigSchema),
    )

    with mock.patch(
        "requests.Session.post",
        autospec=True,
        side_effect=mocked_datasource_post_response,
    ) as mock_post, mock.patch(
        "requests.Session.get",
        autospec=True,
        side_effect=mocked_datasource_get_response,
    ):

        retrieved_datasource_config = store.set(key=None, value=datasource_config)

        serializer = NamedDatasourceSerializer(schema=datasourceConfigSchema)
        expected_datasource_config = serializer.serialize(datasource_config)

        mock_post.assert_called_with(
            mock.ANY,  # requests.Session object
            f"{ge_cloud_base_url}/organizations/{ge_cloud_organization_id}/datasources",
            json={
                "data": {
                    "type": "datasource",
                    "attributes": {
                        "datasource_config": expected_datasource_config,
                        "organization_id": ge_cloud_organization_id,
                    },
                }
            },
        )

        json_serializer = JsonDatasourceConfigSerializer(schema=datasourceConfigSchema)

        assert json_serializer.serialize(
            retrieved_datasource_config
        ) == json_serializer.serialize(datasource_config_with_names_and_ids)


@pytest.mark.integration
def test_datasource_store_with_inline_store_backend(
    datasource_config: DatasourceConfig, empty_data_context: DataContext
) -> None:
    inline_store_backend_config: dict = {
        "class_name": "InlineStoreBackend",
        "resource_type": DataContextVariableSchema.DATASOURCES,
        "data_context": empty_data_context,
        "suppress_store_backend_id": True,
    }

    store = DatasourceStore(
        store_name="my_datasource_store",
        store_backend=inline_store_backend_config,
        serializer=YAMLReadyDictDatasourceConfigSerializer(
            schema=datasourceConfigSchema
        ),
    )

    key = DataContextVariableKey(
        resource_name="my_datasource",
    )

    store.set(key=key, value=datasource_config)
    res: DatasourceConfig = store.get(key=key)

    assert isinstance(res, DatasourceConfig)
    set_config_serializer = DictConfigSerializer(schema=datasourceConfigSchema)
    retrieved_config_serializer = YAMLReadyDictDatasourceConfigSerializer(
        schema=datasourceConfigSchema
    )
    _assert_serialized_datasource_configs_are_equal(
        [datasource_config, res], [set_config_serializer, retrieved_config_serializer]
    )


@pytest.mark.unit
def test_datasource_store_add_by_name(
    empty_datasource_store: DatasourceStore,
    datasource_config: DatasourceConfig,
    fake_datasource_name,
) -> None:
    assert len(empty_datasource_store.list_keys()) == 0

    empty_datasource_store.add_by_name(
        datasource_name=fake_datasource_name, datasource_config=datasource_config
    )

    assert len(empty_datasource_store.list_keys()) == 1


@pytest.mark.unit
def test_datasource_store_set(
    empty_datasource_store: DatasourceStore,
    datasource_config: DatasourceConfig,
    fake_datasource_name,
) -> None:
    assert len(empty_datasource_store.list_keys()) == 0

    datasource_config.name = fake_datasource_name
    retrieved_datasource_config: DatasourceConfig = empty_datasource_store.set(
        key=None, value=datasource_config
    )

    assert len(empty_datasource_store.list_keys()) == 1

    # Use a consistent serializer to check equality
    serializer = JsonDatasourceConfigSerializer(schema=datasourceConfigSchema)
    assert serializer.serialize(retrieved_datasource_config) == serializer.serialize(
        datasource_config
    )


@pytest.mark.integration
def test_datasource_store_retrieve_by_name(
    fake_datasource_name,
    datasource_config: DatasourceConfig,
    datasource_store_with_single_datasource: DatasourceStore,
) -> None:
    actual_config: DatasourceConfig = (
        datasource_store_with_single_datasource.retrieve_by_name(
            datasource_name=fake_datasource_name
        )
    )
    set_config_serializer = DictConfigSerializer(schema=datasourceConfigSchema)
    retrieved_config_serializer = YAMLReadyDictDatasourceConfigSerializer(
        schema=datasourceConfigSchema
    )
    _assert_serialized_datasource_configs_are_equal(
        [datasource_config, actual_config],
        [set_config_serializer, retrieved_config_serializer],
    )


@pytest.mark.unit
def test_datasource_store_delete(
    datasource_config: DatasourceConfig,
    datasource_store_with_single_datasource: DatasourceStore,
) -> None:
    initial_keys = datasource_store_with_single_datasource.list_keys()
    assert len(initial_keys) == 1

    datasource_name = initial_keys[0][0]
    datasource_config.name = datasource_name

    datasource_store_with_single_datasource.delete(
        datasource_config=datasource_config,
    )

    assert len(datasource_store_with_single_datasource.list_keys()) == 0


@pytest.mark.integration
def test_datasource_store_update_by_name(
    fake_datasource_name,
    datasource_config: DatasourceConfig,
    datasource_store_with_single_datasource: DatasourceStore,
) -> None:
    updated_base_directory: str = "foo/bar/baz"

    updated_datasource_config = copy.deepcopy(datasource_config)
    updated_datasource_config.data_connectors["tripdata_monthly_configured"][
        "base_directory"
    ] = updated_base_directory

    datasource_store_with_single_datasource.update_by_name(
        datasource_name=fake_datasource_name,
        datasource_config=updated_datasource_config,
    )

    key = DataContextVariableKey(
        resource_name=fake_datasource_name,
    )
    actual_config = cast(
        DatasourceConfig, datasource_store_with_single_datasource.get(key=key)
    )

    set_config_serializer = DictConfigSerializer(schema=datasourceConfigSchema)
    retrieved_config_serializer = YAMLReadyDictDatasourceConfigSerializer(
        schema=datasourceConfigSchema
    )
    _assert_serialized_datasource_configs_are_equal(
        [updated_datasource_config, actual_config],
        [set_config_serializer, retrieved_config_serializer],
    )


@pytest.mark.unit
def test_datasource_store_update_raises_error_if_datasource_doesnt_exist(
    fake_datasource_name,
    empty_datasource_store: DatasourceStore,
) -> None:
    updated_datasource_config = DatasourceConfig()
    with pytest.raises(gx_exceptions.DatasourceNotFoundError) as e:
        empty_datasource_store.update_by_name(
            datasource_name=fake_datasource_name,
            datasource_config=updated_datasource_config,
        )

    assert (
        f"Could not find an existing Datasource named {fake_datasource_name}."
        in str(e.value)
    )


@pytest.mark.integration
def test_datasource_store_with_inline_store_backend_config_with_names_does_not_store_datasource_name(
    datasource_config_with_names: DatasourceConfig,
    datasource_config: DatasourceConfig,
    empty_data_context: DataContext,
) -> None:
    inline_store_backend_config: dict = {
        "class_name": "InlineStoreBackend",
        "resource_type": DataContextVariableSchema.DATASOURCES,
        "data_context": empty_data_context,
        "suppress_store_backend_id": True,
    }

    store = DatasourceStore(
        store_name="my_datasource_store",
        store_backend=inline_store_backend_config,
        serializer=YAMLReadyDictDatasourceConfigSerializer(
            schema=datasourceConfigSchema
        ),
    )

    key = DataContextVariableKey(
        resource_name="my_datasource",
    )

    store.set(key=key, value=datasource_config_with_names)
    res: DatasourceConfig = store.get(key=key)

    assert isinstance(res, DatasourceConfig)
    set_config_serializer = DictConfigSerializer(schema=datasourceConfigSchema)
    retrieved_config_serializer = YAMLReadyDictDatasourceConfigSerializer(
        schema=datasourceConfigSchema
    )
    _assert_serialized_datasource_configs_are_equal(
        [datasource_config, res], [set_config_serializer, retrieved_config_serializer]
    )

    with open(
        pathlib.Path(empty_data_context.root_directory) / "great_expectations.yml"
    ) as f:
        context_config_from_disk: dict = yaml.load(f)

    assert "name" not in context_config_from_disk["datasources"]["my_datasource"]


@pytest.mark.integration
def test_datasource_store_with_inline_store_backend_config_with_names_does_not_store_dataconnector_name(
    datasource_config_with_names: DatasourceConfig,
    datasource_config: DatasourceConfig,
    empty_data_context: DataContext,
) -> None:
    inline_store_backend_config: dict = {
        "class_name": "InlineStoreBackend",
        "resource_type": DataContextVariableSchema.DATASOURCES,
        "data_context": empty_data_context,
        "suppress_store_backend_id": True,
    }

    store = DatasourceStore(
        store_name="my_datasource_store",
        store_backend=inline_store_backend_config,
        serializer=YAMLReadyDictDatasourceConfigSerializer(
            schema=datasourceConfigSchema
        ),
    )

    key = DataContextVariableKey(
        resource_name="my_datasource",
    )

    store.set(key=key, value=datasource_config_with_names)
    res: DatasourceConfig = store.get(key=key)

    assert isinstance(res, DatasourceConfig)
    set_config_serializer = DictConfigSerializer(schema=datasourceConfigSchema)
    retrieved_config_serializer = YAMLReadyDictDatasourceConfigSerializer(
        schema=datasourceConfigSchema
    )
    _assert_serialized_datasource_configs_are_equal(
        [datasource_config, res], [set_config_serializer, retrieved_config_serializer]
    )

    with open(
        pathlib.Path(empty_data_context.root_directory) / "great_expectations.yml"
    ) as f:
        context_config_from_disk: dict = yaml.load(f)

    assert (
        "name"
        not in context_config_from_disk["datasources"]["my_datasource"][
            "data_connectors"
        ]["tripdata_monthly_configured"]
    )



"""Example Script: How to configure a Pandas Datasource

This example script is intended for use in documentation on how to configure a Pandas Datasource.

Assert statements are included to ensure that if the behaviour shown in this script breaks it will not pass
tests and will be updated.  These statements can be ignored by users.

Comments with the tags `<snippet>` and `</snippet>` are used to ensure that if this script is updated
the snippets that are specified for use in documentation are maintained.  These comments can be ignored by users.

--documentation--
    https://docs.greatexpectations.io/docs/guides/connecting_to_your_data/datasource_configuration/how_to_configure_a_pandas_datasource
"""

# The following imports are used as part of verifying that all example snippets are consistent.
# Users may disregard them.

import operator
from functools import reduce
from typing import List, Tuple

# Import the necessary modules for the examples, and initialize a Data Context.
# <snippet name="filesystem_datasource_config_gx_imports">
from ruamel import yaml

import great_expectations as gx

data_context: gx.DataContext = gx.get_context()
# </snippet>

# The following methods are used to verify that the example configuration snippets are consistent with each other.
# Users can disregard them.


def get_by_path(root_dictionary: dict, keys: Tuple[str]) -> Tuple:
    return "/".join(keys), reduce(operator.getitem, keys, root_dictionary)


def gather_key_paths(
    target_dict: dict, current_path: List[str] = None
) -> List[Tuple[str]]:
    key_paths: List[Tuple[str]] = []
    for key, value in target_dict.items():
        if isinstance(value, dict):
            if current_path:
                next_path = current_path[:]
                next_path.append(key)
            else:
                next_path = [key]
            if value:
                key_paths.extend(gather_key_paths(value, next_path))
            else:
                # If this is an empty dictionary, then there will be no further nested keys to gather.
                key_paths.append(tuple(next_path))
        else:
            if current_path:
                next_path = current_path[:]
                next_path.append(key)
                key_paths.append(tuple(next_path))
            else:
                key_paths.append((key,))
    return key_paths


def is_subset(subset, superset):
    key_paths = gather_key_paths(subset)
    subset_items = [get_by_path(subset, key_path) for key_path in key_paths]
    # If the last value in key_paths leads to an empty placeholder, remove that entry from subset_items
    subset_items = [x for x in subset_items if x[1]]
    # if not subset_items[-1][1]:
    #     subset_items.pop(-1)
    # if not get_by_path(subset, key_paths[-1])[1]:
    #     subset_items.pop(-1)

    superset_items = [get_by_path(superset, key_path) for key_path in key_paths]
    assert all(
        item in superset_items for item in subset_items
    ), f"\n{subset_items} is not a subset of \n{superset_items}"


# The following methods return the full configurations used in the documentation.
# They are defined here to ensure that the partial examples are all consistent with the
# final configuration.


def get_full_pandas_inferred_datasource_single_batch():
    # <snippet name="pandas_inferred_datasource_single_batch_full_snippet">
    datasource_config: dict = {
        "name": "my_datasource_name",  # Preferably name it something relevant
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "PandasExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                "class_name": "InferredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "default_regex": {
                    "pattern": "(.*)\\.csv",
                    "group_names": ["data_asset_name"],
                },
                "batch_spec_passthrough": {
                    "reader_method": "csv",
                    "reader_options": {
                        "header": True,
                        "inferSchema": True,
                    },
                },
            }
        },
    }
    # </snippet>
    return datasource_config


def get_full_pandas_inferred_datasource_multi_batch():
    # <snippet name="pandas_inferred_datasource_multi_batch_full_snippet">
    datasource_config: dict = {
        "name": "my_datasource_name",  # Preferably name it something relevant
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "PandasExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                "class_name": "InferredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "default_regex": {
                    "pattern": "(yellow_tripdata_sample_2020)-(\\d.*)\\.csv",
                    "group_names": ["data_asset_name", "month"],
                },
                "batch_spec_passthrough": {
                    "reader_method": "csv",
                    "reader_options": {
                        "header": True,
                        "inferSchema": True,
                    },
                },
            }
        },
    }
    # </snippet>
    return datasource_config


def get_full_pandas_configured_datasource_single_batch():
    # <snippet name="pandas_configured_datasource_single_batch_full_snippet">
    datasource_config: dict = {
        "name": "my_datasource_name",  # Preferably name it something relevant
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "PandasExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_configured_data_connector": {
                "class_name": "ConfiguredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "assets": {
                    "yellow_tripdata_jan": {
                        "pattern": "yellow_tripdata_sample_2020-(01)\\.csv",
                        "group_names": ["month"],
                    }
                },
                "batch_spec_passthrough": {
                    "reader_method": "csv",
                    "reader_options": {
                        "header": True,
                        "inferSchema": True,
                    },
                },
            }
        },
    }
    # </snippet>
    return datasource_config


def get_full_pandas_configured_datasource_multi_batch():
    # <snippet name="pandas_configured_datasource_multi_batch_full_snippet">
    datasource_config: dict = {
        "name": "my_datasource_name",  # Preferably name it something relevant
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "PandasExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_configured_data_connector": {
                "class_name": "ConfiguredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "assets": {
                    "yellow_tripdata_2020": {
                        "pattern": "yellow_tripdata_sample_2020-(.*)\\.csv",
                        "group_names": ["month"],
                    }
                },
                "batch_spec_passthrough": {
                    "reader_method": "csv",
                    "reader_options": {
                        "header": True,
                        "inferSchema": True,
                    },
                },
            }
        },
    }
    # </snippet>
    return datasource_config


def get_full_pandas_runtime_datasource():
    # <snippet name="pandas_runtime_datasource_full_snippet">
    datasource_config: dict = {
        "name": "my_datasource_name",  # Preferably name it something relevant
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "PandasExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_runtime_data_connector": {
                "class_name": "RuntimeDataConnector",
                "batch_spec_passthrough": {
                    "reader_method": "csv",
                    "reader_options": {
                        "header": True,
                        "inferSchema": True,
                    },
                },
                "batch_identifiers": ["batch_timestamp"],
            }
        },
    }
    # </snippet>
    return datasource_config


def validate_pandas_datasource_configuration_snippets():
    """
    Tests that the configuration snippets for all keys that are required by Inferred, Configured, and Runtime Pandas
    Datasources are consistent with each other and with the full configurations presented in the guide.

    """
    # Snippet: create an empty dict for your configuration.
    # <snippet name="datasource_configuration_empty_dictionary">
    datasource_config: dict = {}
    # </snippet>

    # Snippet: adding a name to your datasource
    datasource_config: dict = {
        # <snippet name="datasource_configuration_name_key">
        "name": "my_datasource_name",  # Preferably name it something relevant
        # </snippet>
    }
    prev_snippet = datasource_config

    # Snippet: full config after adding a name to your Datasource.
    # <snippet name="datasource_configuration_post_name_key">
    datasource_config: dict = {
        "name": "my_datasource_name",  # Preferably name it something relevant
    }
    # </snippet>
    assert datasource_config == prev_snippet

    # Snippet: Adding a class_name and module_name to your Datasource.
    datasource_config: dict = {
        "name": "my_datasource_name",  # Preferably name it something relevant
        # <snippet name="datasource_configuration_class_and_module_keys">
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource"
        # </snippet>
    }
    is_subset(prev_snippet, datasource_config)
    prev_snippet = datasource_config

    # Snippet: Full configuration after adding class_name and module_name to your Datasource.
    # <snippet name="datasource_configuration_post_class_and_module_keys">
    datasource_config: dict = {
        "name": "my_datasource_name",  # Preferably name it something relevant
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
    }
    # </snippet>
    assert datasource_config == prev_snippet

    # Snippet: Add an execution_engine to your Datasource.
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        # <snippet name="datasource_configuration_add_execution_engine">
        "execution_engine": {
            "class_name": "PandasExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        # </snippet>
    }
    is_subset(prev_snippet, datasource_config)
    prev_snippet = datasource_config

    # Snippet: Full config after adding an Execution Engine to your Datasource.
    # <snippet name="datasource_configuration_post_execution_engine">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "PandasExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
    }
    # </snippet>
    assert datasource_config == prev_snippet
    prev_snippet = datasource_config

    # Snippet: Add an empty dictionary for your data_connectors configuration.
    # <snippet name="datasource_configuration_add_empty_data_connectors_key">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "PandasExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {},
    }
    # </snippet>
    is_subset(prev_snippet, datasource_config)

    # Verify that the final version of the universally relevant configuration is consistent with the
    # full configurations demonstrated in the guide.
    for full_config in [
        get_full_pandas_configured_datasource_single_batch(),
        get_full_pandas_inferred_datasource_single_batch(),
        get_full_pandas_configured_datasource_multi_batch(),
        get_full_pandas_inferred_datasource_multi_batch(),
        get_full_pandas_runtime_datasource(),
    ]:
        is_subset(datasource_config, full_config)


def validate_pandas_datasource_configuration_inferred_snippets():
    # Get the full configurations to test against
    full_inferred_single_batch_config = (
        get_full_pandas_inferred_datasource_single_batch()
    )
    full_inferred_multi_batch_config = get_full_pandas_inferred_datasource_multi_batch()

    # Snippet for adding the Data Connector configuration dictionary.
    # <snippet name="datasource_config add empty inferred_data_connector">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "PandasExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {"name_of_my_inferred_data_connector": {}},
    }
    # </snippet>
    is_subset(datasource_config, full_inferred_single_batch_config)
    is_subset(datasource_config, full_inferred_multi_batch_config)

    # Snippet for populating the Data Connector class_name
    datasource_config: dict = {
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                # <snippet name="inferred data connector populate class_name">
                "class_name": "InferredAssetFilesystemDataConnector",
                # </snippet>
            }
        }
    }
    is_subset(datasource_config, full_inferred_single_batch_config)
    is_subset(datasource_config, full_inferred_multi_batch_config)

    # Snippet for adding the Data Connector base_directory
    datasource_config: dict = {
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                # <snippet name="inferred data connector add base_directory">
                "base_directory": "../data",
                # </snippet>
            }
        }
    }
    is_subset(datasource_config, full_inferred_single_batch_config)
    is_subset(datasource_config, full_inferred_multi_batch_config)

    # Snippet for the final version of data_connectors class_name and module_name configuration.
    # <snippet name="datasource_configuration_pandas_inferred_empty_regex">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "PandasExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                "class_name": "InferredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "default_regex": {},
                "batch_spec_passthrough": {},
            }
        },
    }
    # </snippet>
    is_subset(datasource_config, full_inferred_single_batch_config)
    is_subset(datasource_config, full_inferred_multi_batch_config)

    # Snippet: Add a glob_directive
    datasource_config = {
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                # <snippet name="datasource_configuration_glob_directive">
                "glob_directive": "*.*"
                # </snippet>
            }
        }
    }
    prev_config = datasource_config
    # Snippet: Full config after adding glob_directive.
    # <snippet name="datasource_configuration_post_glob_directive">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "PandasExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                "class_name": "InferredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "glob_directive": "*.*",
                "default_regex": {},
            }
        },
    }
    # </snippet>

    # Verify that the two glob_directive snippets are consistent with each other.
    is_subset(prev_config, datasource_config)
    # Verify that the final glob_directive snippet is consistent with the full configs, other than the "glob_directive"
    # key which is not included in the final configuration.
    del datasource_config["data_connectors"]["name_of_my_inferred_data_connector"][
        "glob_directive"
    ]
    is_subset(datasource_config, full_inferred_single_batch_config)
    is_subset(datasource_config, full_inferred_multi_batch_config)


def validate_pandas_datasource_configuration_inferred_single_batch_snippets():
    full_inferred_single_batch_config = (
        get_full_pandas_inferred_datasource_single_batch()
    )
    # Snippet: single batch pattern in default_regex
    # Snippet: single batch group_names in default_regex
    datasource_config = {
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                "default_regex": {
                    # <snippet name="datasource_configuration_inferred_single_batch_regex_pattern">
                    "pattern": "(.*)\\.csv",
                    # </snippet>
                    # <snippet name="datasource_configuration_inferred_single_batch_group_names">
                    "group_names": ["data_asset_name"],
                    # </snippet>
                }
            }
        }
    }
    is_subset(datasource_config, full_inferred_single_batch_config)

    # Snippet: full inferred_data_connector with single batch default_regex
    datasource_config = {
        "data_connectors": {
            # <snippet name="data_connector_configuration_post_single_batch_regex">
            "name_of_my_inferred_data_connector": {
                "class_name": "InferredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "default_regex": {
                    "pattern": "(.*)\\.csv",
                    "group_names": ["data_asset_name"],
                },
            }
            # </snippet>
        }
    }
    is_subset(datasource_config, full_inferred_single_batch_config)

    # Snippet: Full configuration for inferred Datasource with single batch default_regex.
    # <snippet name="datasource_configuration_post_inferred_data_connector">
    datasource_config = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "PandasExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                "class_name": "InferredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "default_regex": {
                    "pattern": "(.*)\\.csv",
                    "group_names": ["data_asset_name"],
                },
            }
        },
    }
    # </snippet>
    is_subset(datasource_config, full_inferred_single_batch_config)


def validate_pandas_datasource_configuration_inferred_multi_batch_snippets():
    full_inferred_multi_batch_config = get_full_pandas_inferred_datasource_multi_batch()
    # Snippet: multibatch pattern for inferred Data Connector default_regex
    # Snippet: multibatch group_names for inferred Data Connector default_regex
    datasource_config = {
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                "default_regex": {
                    # <snippet name="datasource_configuration_inferred_multi_batch_regex_pattern">
                    "pattern": "(yellow_tripdata_sample_2020)-(\\d.*)\\.csv",
                    # </snippet>
                    # <snippet name="group_names for inferred multi batch default_regex key">
                    "group_names": ["data_asset_name", "month"],
                    # </snippet>
                }
            }
        }
    }
    is_subset(datasource_config, full_inferred_multi_batch_config)

    # Snippet: full data connector config for inferred Data Connector multi batch
    datasource_config = {
        "data_connectors": {
            # <snippet name="data connector config for inferred filesystem data connector multi batch">
            "name_of_my_inferred_data_connector": {
                "class_name": "InferredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "default_regex": {
                    "pattern": "(yellow_tripdata_sample_2020)-(\\d.*)\\.csv",
                    "group_names": ["data_asset_name", "month"],
                },
            }
            # </snippet>
        }
    }
    is_subset(datasource_config, full_inferred_multi_batch_config)

    # Snippet: Full configuration for Pandas multibatch Inferred Datasource.
    # <snippet name="datasource_configuration_post_multi_batch_inferred_data_connector">
    datasource_config = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "PandasExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                "class_name": "InferredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "default_regex": {
                    "pattern": "(yellow_tripdata_sample_2020)-(\\d.*)\\.csv",
                    "group_names": ["data_asset_name", "month"],
                },
            }
        },
    }
    # </snippet>
    is_subset(datasource_config, full_inferred_multi_batch_config)


def validate_pandas_datasource_configuration_configured_snippets():
    # Get the full configurations to test against
    full_configured_single_batch_config = (
        get_full_pandas_configured_datasource_single_batch()
    )
    full_configured_multi_batch_config = (
        get_full_pandas_configured_datasource_multi_batch()
    )

    # Snippet: Add a dictionary for your configured data_connector
    # <snippet name="datasource_configuration_add_empty_configured_data_connector">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "PandasExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {"name_of_my_configured_data_connector": {}},
    }
    # </snippet>
    is_subset(datasource_config, full_configured_single_batch_config)
    is_subset(datasource_config, full_configured_multi_batch_config)

    # Snippet: Add class_name to your data connector config.
    datasource_config: dict = {
        "data_connectors": {
            "name_of_my_configured_data_connector": {
                # <snippet name="data_connector_configuration_configured_class_name">
                "class_name": "ConfiguredAssetFilesystemDataConnector",
                # </snippet>
                # <snippet name="data_connector_configuration_configured_base_directory">
                "base_directory": "../data",
                # </snippet>
            }
        }
    }
    is_subset(datasource_config, full_configured_single_batch_config)
    is_subset(datasource_config, full_configured_multi_batch_config)

    # Snippet: Full config for data connector, with empty assets dictionary
    # <snippet name="datasource_config_configured_data_connector_empty_data_asset">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "PandasExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_configured_data_connector": {
                "class_name": "ConfiguredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "assets": {},
                "batch_spec_passthrough": {},
            }
        },
    }
    # </snippet>
    is_subset(datasource_config, full_configured_single_batch_config)
    is_subset(datasource_config, full_configured_multi_batch_config)


def validate_pandas_datasource_configuration_configured_single_batch_snippets():
    # Get the full configurations to test against
    full_configured_single_batch_config = (
        get_full_pandas_configured_datasource_single_batch()
    )

    # Snippet: Empty dictionary for an asset configuration.
    datasource_config = {
        "data_connectors": {
            "name_of_my_configured_data_connector": {
                "assets": {
                    # <snippet name="empty_data_asset_configuration_configured_single_batch">
                    "yellow_tripdata_jan": {}
                    # </snippet>
                }
            }
        }
    }
    is_subset(datasource_config, full_configured_single_batch_config)

    # Snippet: Single batch pattern for an asset configuration.
    # Snippet: Single batch group_names for an asset configuration.
    datasource_config = {
        "data_connectors": {
            "name_of_my_configured_data_connector": {
                "assets": {
                    "yellow_tripdata_jan": {
                        # <snippet name="pattern_for_single_batch_configured_data_asset_configuration">
                        "pattern": "yellow_tripdata_sample_2020-(01)\\.csv",
                        # </snippet>
                        # <snippet name="group_names for single batch configured assets configuration">
                        "group_names": ["month"],
                        # </snippet>
                    }
                }
            }
        }
    }
    is_subset(datasource_config, full_configured_single_batch_config)

    # Snippet: Full single batch asset configuration.
    datasource_config = {
        "data_connectors": {
            "name_of_my_configured_data_connector": {
                "assets": {
                    # <snippet name="full_data_asset_for_single_batch_configured_data_connector">
                    "yellow_tripdata_jan": {
                        "pattern": "yellow_tripdata_sample_2020-(01)\\.csv",
                        "group_names": ["month"],
                    }
                    # </snippet>
                }
            }
        }
    }
    is_subset(datasource_config, full_configured_single_batch_config)

    # Snippet: Full configuration for a Configured Data Connector with a single-batch Data Asset.
    datasource_config = {
        "data_connectors": {
            # <snippet name="full_single_batch_configured_data_connector_configuration">
            "name_of_my_configured_data_connector": {
                "class_name": "ConfiguredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "assets": {
                    "yellow_tripdata_jan": {
                        "pattern": "yellow_tripdata_sample_2020-(01)\\.csv",
                        "group_names": ["month"],
                    }
                },
            }
            # </snippet>
        }
    }
    is_subset(datasource_config, full_configured_single_batch_config)

    # Snippet: Full configuration for a Datasource using a Configured Data Connector and single-batch Data Asset.
    # <snippet name="full datasource_config for pandas execution engine, configured data connector; single batch asset">
    datasource_config = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "PandasExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_configured_data_connector": {
                "class_name": "ConfiguredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "assets": {
                    "yellow_tripdata_jan": {
                        "pattern": "yellow_tripdata_sample_2020-(01)\\.csv",
                        "group_names": ["month"],
                    }
                },
            }
        },
    }
    # </snippet>
    is_subset(datasource_config, full_configured_single_batch_config)


def validate_pandas_datasource_configuration_configured_multi_batch_snippets():
    # Get the full configurations to test against
    full_configured_multi_batch_config = (
        get_full_pandas_configured_datasource_multi_batch()
    )

    # Snippet: Empty dictionary for an asset configuration.
    datasource_config = {
        "data_connectors": {
            "name_of_my_configured_data_connector": {
                "assets": {
                    # <snippet name="empty data asset for configured, multi batch datasource">
                    "yellow_tripdata_2020": {}
                    # </snippet>
                }
            }
        }
    }
    is_subset(datasource_config, full_configured_multi_batch_config)

    # Snippet: Single batch pattern for an asset configuration.
    # Snippet: Single batch group_names for an asset configuration.
    datasource_config = {
        "data_connectors": {
            "name_of_my_configured_data_connector": {
                "assets": {
                    "yellow_tripdata_2020": {
                        # <snippet name="pattern for asset in configured, multi-batch datasource">
                        "pattern": "yellow_tripdata_sample_2020-(.*)\\.csv",
                        # </snippet>
                        # <snippet name="group_names for asset in configured, multi-batch datasource">
                        "group_names": ["month"],
                        # </snippet>
                    }
                }
            }
        }
    }
    is_subset(datasource_config, full_configured_multi_batch_config)

    # Snippet: Full single batch asset configuration.
    datasource_config = {
        "data_connectors": {
            "name_of_my_configured_data_connector": {
                "assets": {
                    # <snippet name="multi batch data asset for configured pandas datasource">
                    "yellow_tripdata_2020": {
                        "pattern": "yellow_tripdata_sample_2020-(.*)\\.csv",
                        "group_names": ["month"],
                    }
                    # </snippet>
                }
            }
        }
    }
    is_subset(datasource_config, full_configured_multi_batch_config)

    # Snippet: Full configuration for a Configured Data Connector with a single-batch Data Asset.
    datasource_config = {
        "data_connectors": {
            # <snippet name="data connector for configured, multi-batch, Pandas datasource">
            "name_of_my_configured_data_connector": {
                "class_name": "ConfiguredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "assets": {
                    "yellow_tripdata_2020": {
                        "pattern": "yellow_tripdata_sample_2020-(.*)\\.csv",
                        "group_names": ["month"],
                    }
                },
            }
            # </snippet>
        }
    }
    is_subset(datasource_config, full_configured_multi_batch_config)

    # Snippet: Full configuration for a Datasource using a Configured Data Connector and single-batch Data Asset.
    # <snippet name="full config for Pandas, configured, multi-batch Datasource">
    datasource_config = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "PandasExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_configured_data_connector": {
                "class_name": "ConfiguredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "assets": {
                    "yellow_tripdata_2020": {
                        "pattern": "yellow_tripdata_sample_2020-(.*)\\.csv",
                        "group_names": ["month"],
                    }
                },
            }
        },
    }
    # </snippet>
    is_subset(datasource_config, full_configured_multi_batch_config)


def validate_pandas_datsource_configuration_runtime_snippets():
    full_runtime_config = get_full_pandas_runtime_datasource()

    # Snippet: Add empty dictionary for runtime data connector.
    # <snippet name="add empty dictionary for runtime data connector">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "PandasExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {"name_of_my_runtime_data_connector": {}},
    }
    # </snippet>
    is_subset(datasource_config, full_runtime_config)

    # Snippet: add class name to runtime Data Connector config.
    datasource_config: dict = {
        "data_connectors": {
            "name_of_my_runtime_data_connector": {
                # <snippet name="class_name for pandas runtime data connector">
                "class_name": "RuntimeDataConnector",
                # </snippet>
            }
        }
    }
    is_subset(datasource_config, full_runtime_config)

    # Snippet: Full datasource config for runtime data connector with blank batch_identifiers list.
    # <snippet name="full config for pandas runtime Datasource with blank data asset identifiers">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "PandasExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_runtime_data_connector": {
                "class_name": "RuntimeDataConnector",
                "batch_spec_passthrough": {},
                "batch_identifiers": [],
            }
        },
    }
    # </snippet>
    is_subset(datasource_config, full_runtime_config)

    # Snippet: Adding a batch_identifiers entry for batch_timestamp.
    datasource_config: dict = {
        "data_connectors": {
            "name_of_my_runtime_data_connector": {
                # <snippet name="batch_identifiers for filesystem runtime data connector">
                "batch_identifiers": ["batch_timestamp"]
                # </snippet>
            }
        }
    }
    is_subset(datasource_config, full_runtime_config)

    # Snippet: Full Pandas Datasource configuration with Runtime data connector and timestamp batch identifier.
    # <snippet name="full pandas runtime Datasource configuration">
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "PandasExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_runtime_data_connector": {
                "class_name": "RuntimeDataConnector",
                "batch_spec_passthrough": {
                    "reader_method": "csv",
                    "reader_options": {
                        "header": True,
                        "inferSchema": True,
                    },
                },
                "batch_identifiers": ["batch_timestamp"],
            }
        },
    }
    # </snippet>
    is_subset(datasource_config, full_runtime_config)


def validate_pandas_batch_spec_passthrough_config_for_inferred():
    full_configs = (
        get_full_pandas_inferred_datasource_single_batch(),
        get_full_pandas_inferred_datasource_multi_batch(),
        # get_full_pandas_configured_datasource_single_batch(),
        # get_full_pandas_configured_datasource_multi_batch(),
        # get_full_pandas_runtime_datasource(),
    )

    datasource_config: dict = {
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                # <snippet name="empty batch_spec_passthrough">
                "batch_spec_passthrough": {
                    "reader_method": "",
                    "reader_options": {},
                    # </snippet>
                },
            }
        },
    }
    for full_config in full_configs:
        is_subset(datasource_config, full_config)

    datasource_config: dict = {
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                "batch_spec_passthrough": {
                    # <snippet name="set reader_method to csv">
                    "reader_method": "csv",
                    # </snippet>
                    # <snippet name="empty keys for reader_options">
                    "reader_options": {
                        "header": "",
                        "inferSchema": "",
                    },
                    # </snippet>
                },
            }
        },
    }
    for full_config in full_configs:
        is_subset(datasource_config, full_config)

    datasource_config: dict = {
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                "batch_spec_passthrough": {
                    "reader_options": {
                        # <snippet name="populate header for reader_options as True">
                        "header": True,
                        # </snippet>
                        # <snippet name="populate inferSchema for reader_options as True">
                        "inferSchema": True,
                        # </snippet>
                    },
                },
            }
        },
    }
    for full_config in full_configs:
        is_subset(datasource_config, full_config)

    datasource_config: dict = {
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                # <snippet name="fully populated batch_spec_passthrough configuration">
                "batch_spec_passthrough": {
                    "reader_method": "csv",
                    "reader_options": {
                        "header": True,
                        "inferSchema": True,
                    },
                },
                # </snippet>
            }
        },
    }
    for full_config in full_configs:
        is_subset(datasource_config, full_config)

    # <snippet name="inferred datasource_config up to batch_spec_passthrough">
    datasource_config: dict = {
        "name": "my_datasource_name",  # Preferably name it something relevant
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "PandasExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                "class_name": "InferredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "default_regex": {},
                "batch_spec_passthrough": {
                    "reader_method": "csv",
                    "reader_options": {
                        "header": True,
                        "inferSchema": True,
                    },
                },
            }
        },
    }
    # </snippet>
    is_subset(datasource_config, get_full_pandas_inferred_datasource_multi_batch())
    is_subset(datasource_config, get_full_pandas_inferred_datasource_single_batch())

    # <snippet name="configured datasource_config up to batch_spec_passthrough">
    datasource_config: dict = {
        "name": "my_datasource_name",  # Preferably name it something relevant
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "PandasExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_configured_data_connector": {
                "class_name": "ConfiguredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "batch_spec_passthrough": {
                    "reader_method": "csv",
                    "reader_options": {
                        "header": True,
                        "inferSchema": True,
                    },
                },
            }
        },
    }
    # </snippet>
    is_subset(datasource_config, get_full_pandas_configured_datasource_multi_batch())
    is_subset(datasource_config, get_full_pandas_configured_datasource_single_batch())

    # <snippet name="runtime datasource_config up to batch_spec_passthrough">
    datasource_config: dict = {
        "name": "my_datasource_name",  # Preferably name it something relevant
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "PandasExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_runtime_data_connector": {
                "class_name": "RuntimeDataConnector",
                "batch_spec_passthrough": {
                    "reader_method": "csv",
                    "reader_options": {
                        "header": True,
                        "inferSchema": True,
                    },
                },
                "batch_identifiers": [],
            }
        },
    }
    # </snippet>
    is_subset(datasource_config, get_full_pandas_runtime_datasource())


validate_pandas_datasource_configuration_snippets()

validate_pandas_datasource_configuration_inferred_snippets()
validate_pandas_datasource_configuration_inferred_single_batch_snippets()
validate_pandas_datasource_configuration_inferred_multi_batch_snippets()

validate_pandas_datasource_configuration_configured_snippets()
validate_pandas_datasource_configuration_configured_single_batch_snippets()
validate_pandas_datasource_configuration_configured_multi_batch_snippets()

validate_pandas_datsource_configuration_runtime_snippets()

validate_pandas_batch_spec_passthrough_config_for_inferred()


def test_pandas_inferred_single_batch_full_configuration():
    datasource_config = get_full_pandas_inferred_datasource_single_batch()
    test_result = data_context.test_yaml_config(yaml.dump(datasource_config))
    datasource_check = test_result.self_check(max_examples=12)

    # NOTE: The following code is only for testing and can be ignored by users.
    # Assert that all of our files have individually become data assets.
    assert (
        datasource_check["data_connectors"]["name_of_my_inferred_data_connector"][
            "data_asset_count"
        ]
        == 12
    )
    # Assert that all of our data assets have only a single batch definition.
    for data_asset in datasource_check["data_connectors"][
        "name_of_my_inferred_data_connector"
    ]["data_assets"].values():
        assert data_asset["batch_definition_count"] == 1
    # Assert that all of our files are now associated with data assets.
    assert (
        datasource_check["data_connectors"]["name_of_my_inferred_data_connector"][
            "unmatched_data_reference_count"
        ]
        == 0
    )
    # Assert that the data assets have the expected name format.
    assert "yellow_tripdata_sample_2020-01" in set(
        datasource_check["data_connectors"]["name_of_my_inferred_data_connector"][
            "data_assets"
        ].keys()
    )
    # Assert that the data assets' data references consist of a list of a single .csv file.
    assert ["yellow_tripdata_sample_2020-01.csv",] in [
        _["example_data_references"]
        for _ in datasource_check["data_connectors"][
            "name_of_my_inferred_data_connector"
        ]["data_assets"].values()
    ]


def test_pandas_inferred_multi_batch_full_configuration():
    datasource_config = get_full_pandas_inferred_datasource_multi_batch()
    test_result = data_context.test_yaml_config(yaml.dump(datasource_config))
    datasource_check = test_result.self_check(max_examples=12)

    # NOTE: The following code is only for testing and can be ignored by users.
    # Assert that all of our files have individually become data assets.
    assert (
        datasource_check["data_connectors"]["name_of_my_inferred_data_connector"][
            "data_asset_count"
        ]
        == 1
    )
    # Assert that our data asset contains all of the source files.
    assert (
        datasource_check["data_connectors"]["name_of_my_inferred_data_connector"][
            "data_assets"
        ]["yellow_tripdata_sample_2020"]["batch_definition_count"]
        == 12
    )
    # Assert that all of our files are now associated with data assets.
    assert (
        datasource_check["data_connectors"]["name_of_my_inferred_data_connector"][
            "unmatched_data_reference_count"
        ]
        == 0
    )


def test_pandas_configured_single_batch_full_configuration():
    datasource_config = get_full_pandas_configured_datasource_single_batch()
    test_result = data_context.test_yaml_config(yaml.dump(datasource_config))
    datasource_check = test_result.self_check(max_examples=12)

    # NOTE: The following code is only for testing and can be ignored by users.
    # Assert that there is only one data asset, since only one was explicitly defined.
    assert (
        datasource_check["data_connectors"]["name_of_my_configured_data_connector"][
            "data_asset_count"
        ]
        == 1
    )
    # Assert that all of our data assets have only a single batch definition.
    for data_asset in datasource_check["data_connectors"][
        "name_of_my_configured_data_connector"
    ]["data_assets"].values():
        assert data_asset["batch_definition_count"] == 1
    # Assert that all of our other files aren't associated with anything.
    assert (
        datasource_check["data_connectors"]["name_of_my_configured_data_connector"][
            "unmatched_data_reference_count"
        ]
        == 11
    )
    # Assert that the data assets have the expected name format.
    assert (
        "yellow_tripdata_jan"
        in datasource_check["data_connectors"]["name_of_my_configured_data_connector"][
            "data_assets"
        ].keys()
    )
    # Assert that the data assets' data references consist of a list of a single .csv file.
    assert ["yellow_tripdata_sample_2020-01.csv",] in [
        _["example_data_references"]
        for _ in datasource_check["data_connectors"][
            "name_of_my_configured_data_connector"
        ]["data_assets"].values()
    ]


def test_pandas_configured_multi_batch_full_configuration():
    datasource_config = get_full_pandas_configured_datasource_multi_batch()
    test_result = data_context.test_yaml_config(yaml.dump(datasource_config))
    datasource_check = test_result.self_check(max_examples=12)

    # NOTE: The following code is only for testing and can be ignored by users.
    # Assert that all of our files have individually become data assets.
    assert (
        datasource_check["data_connectors"]["name_of_my_configured_data_connector"][
            "data_asset_count"
        ]
        == 1
    )
    # Assert that our data asset contains all of the source files.
    assert (
        datasource_check["data_connectors"]["name_of_my_configured_data_connector"][
            "data_assets"
        ]["yellow_tripdata_2020"]["batch_definition_count"]
        == 12
    )
    # Assert that all of our files are now associated with data assets.
    assert (
        datasource_check["data_connectors"]["name_of_my_configured_data_connector"][
            "unmatched_data_reference_count"
        ]
        == 0
    )


def test_pandas_configured_runtime_full_configuration():
    datasource_config = get_full_pandas_runtime_datasource()

    test_result = data_context.test_yaml_config(yaml.dump(datasource_config))
    datasource_check = test_result.self_check(max_examples=12)

    # NOTE: The following code is only for testing and can be ignored by users.
    # Assert that there are no data sets -- those get defined in a Batch Request.
    assert (
        datasource_check["data_connectors"]["name_of_my_runtime_data_connector"][
            "data_asset_count"
        ]
        == 0
    )


def test_adding_your_datasource_to_the_datacontext():
    datasource_config = get_full_pandas_runtime_datasource()

    # <snippet name="test your config with test_yaml_config">
    data_context.test_yaml_config(yaml.dump(datasource_config))
    # </snippet>

    # <snippet name="add your datasource to your data_context">
    data_context.add_datasource(**datasource_config)
    # </snippet>

    # <snippet name="add your datasource to your data_context only if it does not already exit">
    # add_datasource only if it doesn't already exist in your Data Context
    try:
        data_context.get_datasource(datasource_config["name"])
    except ValueError:
        data_context.add_datasource(**datasource_config)
    else:
        print(
            f"The datasource {datasource_config['name']} already exists in your Data Context!"
        )
    # </snippet>


test_pandas_inferred_single_batch_full_configuration()
test_pandas_inferred_multi_batch_full_configuration()
test_pandas_configured_single_batch_full_configuration()
test_pandas_configured_multi_batch_full_configuration()
test_pandas_configured_runtime_full_configuration()
test_adding_your_datasource_to_the_datacontext()



from __future__ import annotations

import json
import logging
import pathlib
from io import StringIO
from pprint import pformat as pf
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Callable,
    Mapping,
    Type,
    TypeVar,
    Union,
    overload,
)

from ruamel.yaml import YAML

from great_expectations.compatibility import pydantic
from great_expectations.experimental.datasources.constants import _FIELDS_ALWAYS_SET

if TYPE_CHECKING:
    MappingIntStrAny = Mapping[Union[int, str], Any]
    AbstractSetIntStr = AbstractSet[Union[int, str]]

logger = logging.getLogger(__name__)

yaml = YAML(typ="safe")
# NOTE (kilo59): the following settings appear to be what we use in existing codebase
yaml.indent(mapping=2, sequence=4, offset=2)
yaml.default_flow_style = False

# TODO (kilo59): replace this with `typing_extensions.Self` once mypy supports it
# Taken from this SO answer https://stackoverflow.com/a/72182814/6304433
_Self = TypeVar("_Self", bound="ExperimentalBaseModel")


class ExperimentalBaseModel(pydantic.BaseModel):
    """
    Base model for most ZEP pydantic models.

    Adds yaml dumping and parsing methods.

    Extra fields are not allowed.

    Serialization methods default to `exclude_unset = True` to prevent serializing
    configs full of mostly unset default values.
    Also prevents passing along unset kwargs to BatchSpec.
    https://docs.pydantic.dev/usage/exporting_models/
    """

    class Config:
        extra = pydantic.Extra.forbid

    @classmethod
    def parse_yaml(cls: Type[_Self], f: Union[pathlib.Path, str]) -> _Self:
        loaded = yaml.load(f)
        logger.debug(f"loaded from yaml ->\n{pf(loaded, depth=3)}\n")
        # noinspection PyArgumentList
        config = cls(**loaded)
        return config

    @overload
    def yaml(
        self,
        stream_or_path: Union[StringIO, None] = None,
        *,
        include: Union[AbstractSetIntStr, MappingIntStrAny, None] = ...,
        exclude: Union[AbstractSetIntStr, MappingIntStrAny, None] = ...,
        by_alias: bool = ...,
        exclude_unset: bool = ...,
        exclude_defaults: bool = ...,
        exclude_none: bool = ...,
        encoder: Union[Callable[[Any], Any], None] = ...,
        models_as_dict: bool = ...,
        **yaml_kwargs,
    ) -> str:
        ...

    @overload
    def yaml(
        self,
        stream_or_path: pathlib.Path,
        *,
        include: Union[AbstractSetIntStr, MappingIntStrAny, None] = ...,
        exclude: Union[AbstractSetIntStr, MappingIntStrAny, None] = ...,
        by_alias: bool = ...,
        exclude_unset: bool = ...,
        exclude_defaults: bool = ...,
        exclude_none: bool = ...,
        encoder: Union[Callable[[Any], Any], None] = ...,
        models_as_dict: bool = ...,
        **yaml_kwargs,
    ) -> pathlib.Path:
        ...

    def yaml(
        self,
        stream_or_path: Union[StringIO, pathlib.Path, None] = None,
        *,
        include: Union[AbstractSetIntStr, MappingIntStrAny, None] = None,
        exclude: Union[AbstractSetIntStr, MappingIntStrAny, None] = None,
        by_alias: bool = False,
        exclude_unset: bool = True,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        encoder: Union[Callable[[Any], Any], None] = None,
        models_as_dict: bool = True,
        **yaml_kwargs,
    ) -> Union[str, pathlib.Path]:
        """
        Serialize the config object as yaml.

        Writes to a file if a `pathlib.Path` is provided.
        Else it writes to a stream and returns a yaml string.
        """
        if stream_or_path is None:
            stream_or_path = StringIO()

        # pydantic json encoder has support for many more types
        # TODO: can we dump json string directly to yaml.dump?
        intermediate_json = self._json_dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            encoder=encoder,
            models_as_dict=models_as_dict,
        )
        yaml.dump(intermediate_json, stream=stream_or_path, **yaml_kwargs)

        if isinstance(stream_or_path, pathlib.Path):
            return stream_or_path
        return stream_or_path.getvalue()

    def json(
        self,
        *,
        include: AbstractSetIntStr | MappingIntStrAny | None = None,
        exclude: AbstractSetIntStr | MappingIntStrAny | None = None,
        by_alias: bool = False,
        # deprecated - use exclude_unset instead
        skip_defaults: bool | None = None,
        # Default to True to prevent serializing long configs full of unset default values
        exclude_unset: bool = True,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        encoder: Callable[[Any], Any] | None = None,
        models_as_dict: bool = True,
        **dumps_kwargs: Any,
    ) -> str:
        """
        Generate a JSON representation of the model, `include` and `exclude` arguments
        as per `dict()`.

        `encoder` is an optional function to supply as `default` to json.dumps(), other
        arguments as per `json.dumps()`.

        Deviates from pydantic `exclude_unset` `True` by default instead of `False` by
        default.
        """
        self.__fields_set__.update(_FIELDS_ALWAYS_SET)
        return super().json(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            encoder=encoder,
            models_as_dict=models_as_dict,
            **dumps_kwargs,
        )

    def _json_dict(
        self,
        *,
        include: Union[AbstractSetIntStr, MappingIntStrAny, None] = None,
        exclude: Union[AbstractSetIntStr, MappingIntStrAny, None] = None,
        by_alias: bool = False,
        exclude_unset: bool = True,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        encoder: Union[Callable[[Any], Any], None] = None,
        models_as_dict: bool = True,
        **dumps_kwargs,
    ) -> dict:
        """
        JSON compatible dictionary. All complex types removed.
        Prefer `.dict()` or `.json()`
        """
        return json.loads(
            self.json(
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                encoder=encoder,
                models_as_dict=models_as_dict,
                **dumps_kwargs,
            )
        )

    def dict(
        self,
        *,
        include: AbstractSetIntStr | MappingIntStrAny | None = None,
        exclude: AbstractSetIntStr | MappingIntStrAny | None = None,
        by_alias: bool = False,
        # Default to True to prevent serializing long configs full of unset default values
        exclude_unset: bool = True,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        # deprecated - use exclude_unset instead
        skip_defaults: bool | None = None,
    ) -> dict[str, Any]:
        """
        Generate a dictionary representation of the model, optionally specifying which
        fields to include or exclude.

        Deviates from pydantic `exclude_unset` `True` by default instead of `False` by
        default.
        """
        self.__fields_set__.update(_FIELDS_ALWAYS_SET)
        return super().dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            skip_defaults=skip_defaults,
        )

    def __str__(self):
        return self.yaml()



from __future__ import annotations

import logging
import pathlib
import re
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Dict,
    List,
    Optional,
    Type,
    Union,
)

from typing_extensions import Literal

from great_expectations.experimental.datasources.dynamic_pandas import (
    _generate_pandas_data_asset_models,
)
from great_expectations.experimental.datasources.filesystem_data_asset import (
    _FilesystemDataAsset,
)
from great_expectations.experimental.datasources.interfaces import (
    BatchSortersDefinition,
    DataAsset,
    TestConnectionError,
)
from great_expectations.experimental.datasources.pandas_datasource import (
    _PandasDatasource,
)
from great_expectations.experimental.datasources.signatures import _merge_signatures

if TYPE_CHECKING:
    from great_expectations.experimental.datasources.interfaces import BatchSorter

logger = logging.getLogger(__name__)


_FILESYSTEM_READER_METHOD_BLACK_LIST = (
    # "read_csv",
    # "read_json",
    # "read_excel",
    # "read_parquet",
    "read_clipboard",  # not path based
    # "read_feather",
    "read_fwf",  # unhandled type
    "read_gbq",  # not path based
    # "read_hdf",
    # "read_html",
    # "read_orc",
    # "read_pickle",
    # "read_sas",  # invalid json schema
    # "read_spss",
    "read_sql",  # not path based & type-name conflict
    "read_sql_query",  # not path based
    "read_sql_table",  # not path based
    "read_table",  # type-name conflict
    # "read_xml",
)

_FILESYSTEM_ASSET_MODELS = _generate_pandas_data_asset_models(
    _FilesystemDataAsset,
    blacklist=_FILESYSTEM_READER_METHOD_BLACK_LIST,
    use_docstring_from_method=True,
    skip_first_param=True,
)

try:
    # variables only needed for type-hinting
    CSVAsset = _FILESYSTEM_ASSET_MODELS["csv"]
    ExcelAsset = _FILESYSTEM_ASSET_MODELS["excel"]
    JSONAsset = _FILESYSTEM_ASSET_MODELS["json"]
    ORCAsset = _FILESYSTEM_ASSET_MODELS["orc"]
    ParquetAsset = _FILESYSTEM_ASSET_MODELS["parquet"]
except KeyError as key_err:
    logger.info(f"zep - {key_err} asset model could not be generated")
    CSVAsset = _FilesystemDataAsset
    ExcelAsset = _FilesystemDataAsset
    JSONAsset = _FilesystemDataAsset
    ORCAsset = _FilesystemDataAsset
    ParquetAsset = _FilesystemDataAsset


class PandasFilesystemDatasource(_PandasDatasource):
    # class attributes
    asset_types: ClassVar[List[Type[DataAsset]]] = list(
        _FILESYSTEM_ASSET_MODELS.values()
    )

    # instance attributes
    type: Literal["pandas_filesystem"] = "pandas_filesystem"
    name: str
    base_directory: pathlib.Path
    data_context_root_directory: Optional[pathlib.Path] = None
    assets: Dict[str, _FilesystemDataAsset] = {}

    def test_connection(self, test_assets: bool = True) -> None:
        """Test the connection for the PandasDatasource.

        Args:
            test_assets: If assets have been passed to the PandasDatasource, whether to test them as well.

        Raises:
            TestConnectionError: If the connection test fails.
        """
        if not self.base_directory.exists():
            raise TestConnectionError(
                f"Path: {self.base_directory.resolve()} does not exist."
            )

        if self.assets and test_assets:
            for asset in self.assets.values():
                asset.test_connection()

    def add_csv_asset(
        self,
        name: str,
        batching_regex: Optional[Union[re.Pattern, str]] = None,
        glob_directive: str = "**/*",
        order_by: Optional[BatchSortersDefinition] = None,
        **kwargs,  # TODO: update signature to have specific keys & types
    ) -> CSVAsset:  # type: ignore[valid-type]
        """Adds a CSV DataAsst to the present "PandasDatasource" object.

        Args:
            name: The name of the csv asset
            batching_regex: regex pattern that matches csv filenames that is used to label the batches
            glob_directive: glob for selecting files in directory (defaults to `**/*`) or nested directories (e.g. `*/*/*.csv`)
            order_by: sorting directive via either list[BatchSorter] or "{+|-}key" syntax: +/- (a/de)scending; + default
            kwargs: Extra keyword arguments should correspond to ``pandas.read_csv`` keyword args
        """
        batching_regex_pattern: re.Pattern = self.parse_batching_regex_string(
            batching_regex=batching_regex
        )
        order_by_sorters: list[BatchSorter] = self.parse_order_by_sorters(
            order_by=order_by
        )

        asset = CSVAsset(
            name=name,
            batching_regex=batching_regex_pattern,
            glob_directive=glob_directive,
            order_by=order_by_sorters,
            **kwargs,
        )

        return self.add_asset(asset)

    def add_excel_asset(
        self,
        name: str,
        batching_regex: Optional[Union[str, re.Pattern]] = None,
        glob_directive: str = "**/*",
        order_by: Optional[BatchSortersDefinition] = None,
        **kwargs,  # TODO: update signature to have specific keys & types
    ) -> ExcelAsset:  # type: ignore[valid-type]
        """Adds an Excel DataAsst to the present "PandasDatasource" object.

        Args:
            name: The name of the csv asset
            batching_regex: regex pattern that matches csv filenames that is used to label the batches
            glob_directive: glob for selecting files in directory (defaults to `**/*`) or nested directories (e.g. `*/*/*.csv`)
            order_by: sorting directive via either list[BatchSorter] or "{+|-}key" syntax: +/- (a/de)scending; + default
            kwargs: Extra keyword arguments should correspond to ``pandas.read_excel`` keyword args
        """
        batching_regex_pattern: re.Pattern = self.parse_batching_regex_string(
            batching_regex=batching_regex
        )
        order_by_sorters: list[BatchSorter] = self.parse_order_by_sorters(
            order_by=order_by
        )

        asset = ExcelAsset(
            name=name,
            batching_regex=batching_regex_pattern,
            glob_directive=glob_directive,
            order_by=order_by_sorters,
            **kwargs,
        )

        return self.add_asset(asset)

    def add_json_asset(
        self,
        name: str,
        batching_regex: Optional[Union[str, re.Pattern]] = None,
        glob_directive: str = "**/*",
        order_by: Optional[BatchSortersDefinition] = None,
        **kwargs,  # TODO: update signature to have specific keys & types
    ) -> JSONAsset:  # type: ignore[valid-type]
        """Adds a JSON DataAsst to the present "PandasDatasource" object.

        Args:
            name: The name of the csv asset
            batching_regex: regex pattern that matches csv filenames that is used to label the batches
            glob_directive: glob for selecting files in directory (defaults to `**/*`) or nested directories (e.g. `*/*/*.csv`)
            order_by: sorting directive via either list[BatchSorter] or "{+|-}key" syntax: +/- (a/de)scending; + default
            kwargs: Extra keyword arguments should correspond to ``pandas.read_json`` keyword args
        """
        batching_regex_pattern: re.Pattern = self.parse_batching_regex_string(
            batching_regex=batching_regex
        )
        order_by_sorters: list[BatchSorter] = self.parse_order_by_sorters(
            order_by=order_by
        )

        asset = JSONAsset(
            name=name,
            batching_regex=batching_regex_pattern,
            glob_directive=glob_directive,
            order_by=order_by_sorters,
            **kwargs,
        )

        return self.add_asset(asset)

    def add_parquet_asset(
        self,
        name: str,
        batching_regex: Optional[Union[str, re.Pattern]] = None,
        glob_directive: str = "**/*",
        order_by: Optional[BatchSortersDefinition] = None,
        **kwargs,  # TODO: update signature to have specific keys & types
    ) -> ParquetAsset:  # type: ignore[valid-type]
        """Adds a Parquet DataAsst to the present "PandasDatasource" object.

        Args:
            name: The name of the csv asset
            batching_regex: regex pattern that matches csv filenames that is used to label the batches
            glob_directive: glob for selecting files in directory (defaults to `**/*`) or nested directories (e.g. `*/*/*.csv`)
            order_by: sorting directive via either list[BatchSorter] or "{+|-}key" syntax: +/- (a/de)scending; + default
            kwargs: Extra keyword arguments should correspond to ``pandas.read_parquet`` keyword args
        """
        batching_regex_pattern: re.Pattern = self.parse_batching_regex_string(
            batching_regex=batching_regex
        )
        order_by_sorters: list[BatchSorter] = self.parse_order_by_sorters(
            order_by=order_by
        )

        asset = ParquetAsset(
            name=name,
            batching_regex=batching_regex_pattern,
            glob_directive=glob_directive,
            order_by=order_by_sorters,
            **kwargs,
        )

        return self.add_asset(asset)

    # attr-defined issue
    # https://github.com/python/mypy/issues/12472
    add_csv_asset.__signature__ = _merge_signatures(add_csv_asset, CSVAsset, exclude={"type"})  # type: ignore[attr-defined]
    add_excel_asset.__signature__ = _merge_signatures(add_excel_asset, ExcelAsset, exclude={"type"})  # type: ignore[attr-defined]
    add_json_asset.__signature__ = _merge_signatures(add_json_asset, JSONAsset, exclude={"type"})  # type: ignore[attr-defined]
    add_parquet_asset.__signature__ = _merge_signatures(add_parquet_asset, ParquetAsset, exclude={"type"})  # type: ignore[attr-defined]



"""Config Validator for YAML based configs.

This validator evaluates YAML configurations of core Great Expectations components to give feedback on
 whether they have been configured correctly. It is linked to a Data Context since it does update the
 configuration of the Data Context in some cases if the configuration is valid.

 Typical usage example:
 import great_expectations as gx
 context = gx.get_context()
 context.test_yaml_config(my_config)
"""
from __future__ import annotations

import traceback
from typing import TYPE_CHECKING, Any, List, Optional, Tuple, Union, cast

from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap
from typing_extensions import Literal

from great_expectations.alias_types import JSONValues  # noqa: TCH001
from great_expectations.checkpoint import Checkpoint, SimpleCheckpoint
from great_expectations.core.usage_statistics.anonymizers.anonymizer import Anonymizer
from great_expectations.core.usage_statistics.anonymizers.datasource_anonymizer import (
    DatasourceAnonymizer,
)
from great_expectations.core.usage_statistics.usage_statistics import (
    send_usage_message_from_handler,
)
from great_expectations.data_context.store import Store  # noqa: TCH001
from great_expectations.data_context.types.base import (
    CheckpointConfig,
    datasourceConfigSchema,
)
from great_expectations.data_context.util import instantiate_class_from_config
from great_expectations.datasource import DataConnector, Datasource  # noqa: TCH001
from great_expectations.rule_based_profiler import RuleBasedProfiler  # noqa: TCH001
from great_expectations.rule_based_profiler.config import RuleBasedProfilerConfig
from great_expectations.util import filter_properties_dict

if TYPE_CHECKING:
    from great_expectations.data_context import AbstractDataContext


# TODO: check if this can be refactored to use YAMLHandler class
yaml = YAML()
yaml.indent(mapping=2, sequence=4, offset=2)
yaml.default_flow_style = False


class _YamlConfigValidator:
    """Helper class for validating YAML configurations of Great Expectations components.

    Attributes:
        data_context: The subclass of AbstractDataContext used to configure your Great
            Expectations deployment. Stores configuration and is modified by test_yaml_config.
    """

    TEST_YAML_CONFIG_SUPPORTED_STORE_TYPES = [
        "ExpectationsStore",
        "ValidationsStore",
        "HtmlSiteStore",
        "EvaluationParameterStore",
        "MetricStore",
        "SqlAlchemyQueryStore",
        "CheckpointStore",
        "ProfilerStore",
    ]
    TEST_YAML_CONFIG_SUPPORTED_DATASOURCE_TYPES = [
        "Datasource",
        "SimpleSqlalchemyDatasource",
    ]
    TEST_YAML_CONFIG_SUPPORTED_DATA_CONNECTOR_TYPES = [
        "InferredAssetFilesystemDataConnector",
        "ConfiguredAssetFilesystemDataConnector",
        "InferredAssetS3DataConnector",
        "ConfiguredAssetS3DataConnector",
        "InferredAssetAzureDataConnector",
        "ConfiguredAssetAzureDataConnector",
        "InferredAssetGCSDataConnector",
        "ConfiguredAssetGCSDataConnector",
        "InferredAssetSqlDataConnector",
        "ConfiguredAssetSqlDataConnector",
    ]
    TEST_YAML_CONFIG_SUPPORTED_CHECKPOINT_TYPES = [
        "Checkpoint",
        "SimpleCheckpoint",
    ]
    TEST_YAML_CONFIG_SUPPORTED_PROFILER_TYPES = [
        "RuleBasedProfiler",
    ]
    ALL_TEST_YAML_CONFIG_DIAGNOSTIC_INFO_TYPES = [
        "__substitution_error__",
        "__yaml_parse_error__",
        "__custom_subclass_not_core_ge__",
        "__class_name_not_provided__",
    ]
    ALL_TEST_YAML_CONFIG_SUPPORTED_TYPES = (
        TEST_YAML_CONFIG_SUPPORTED_STORE_TYPES
        + TEST_YAML_CONFIG_SUPPORTED_DATASOURCE_TYPES
        + TEST_YAML_CONFIG_SUPPORTED_DATA_CONNECTOR_TYPES
        + TEST_YAML_CONFIG_SUPPORTED_CHECKPOINT_TYPES
        + TEST_YAML_CONFIG_SUPPORTED_PROFILER_TYPES
    )

    def __init__(
        self,
        data_context: AbstractDataContext,
    ):
        """Init _YamlConfigValidator with a Data Context"""
        self._data_context = data_context

    @property
    def usage_statistics_handler(self):
        return self._data_context.usage_statistics_handler

    @property
    def runtime_environment(self):
        return self._data_context.runtime_environment

    @property
    def config_variables(self):
        return self._data_context.config_variables

    def test_yaml_config(  # noqa: C901 - complexity 17
        self,
        yaml_config: str,
        name: Optional[str] = None,
        class_name: Optional[str] = None,
        runtime_environment: Optional[dict] = None,
        pretty_print: bool = True,
        return_mode: Union[
            Literal["instantiated_class"], Literal["report_object"]
        ] = "instantiated_class",
        shorten_tracebacks: bool = False,
    ):
        """Convenience method for testing yaml configs

        test_yaml_config is a convenience method for configuring the moving
        parts of a Great Expectations deployment. It allows you to quickly
        test out configs for system components, especially Datasources,
        Checkpoints, and Stores.

        For many deployments of Great Expectations, these components (plus
        Expectations) are the only ones you'll need.

        test_yaml_config is mainly intended for use within notebooks and tests.

        --Public API--

        --Documentation--
            https://docs.greatexpectations.io/docs/terms/data_context
            https://docs.greatexpectations.io/docs/guides/validation/checkpoints/how_to_configure_a_new_checkpoint_using_test_yaml_config

        Args:
            yaml_config: A string containing the yaml config to be tested
            name: Optional name of the component to instantiate
            class_name: Optional, overridden if provided in the config
            runtime_environment: Optional override for config items
            pretty_print: Determines whether to print human-readable output
            return_mode: Determines what type of object test_yaml_config will return.
                Valid modes are "instantiated_class" and "report_object"
            shorten_tracebacks:If true, catch any errors during instantiation and print only the
                last element of the traceback stack. This can be helpful for
                rapid iteration on configs in a notebook, because it can remove
                the need to scroll up and down a lot.

        Returns:
            The instantiated component (e.g. a Datasource)
            OR
            a json object containing metadata from the component's self_check method.
            The returned object is determined by return_mode.
        """
        if return_mode not in ["instantiated_class", "report_object"]:
            raise ValueError(f"Unknown return_mode: {return_mode}.")

        if runtime_environment is None:
            runtime_environment = {}

        runtime_environment = {
            **runtime_environment,
            **self.runtime_environment,
        }

        usage_stats_event_name: str = "data_context.test_yaml_config"

        # Based on the particular object type we are attempting to instantiate,
        # we may need the original config, the substituted config, or both.
        config = self._test_yaml_config_prepare_config(
            yaml_config=yaml_config, usage_stats_event_name=usage_stats_event_name
        )
        config_with_substitutions = self._test_yaml_config_prepare_substituted_config(
            yaml_config, runtime_environment, usage_stats_event_name
        )

        if "class_name" in config:
            class_name = config["class_name"]

        instantiated_class: Any = None
        usage_stats_event_payload: dict[str, Union[str, List[str]]] = {}

        if pretty_print:
            print("Attempting to instantiate class from config...")
        try:
            if class_name in self.TEST_YAML_CONFIG_SUPPORTED_STORE_TYPES:
                (
                    instantiated_class,
                    usage_stats_event_payload,
                ) = self._test_instantiation_of_store_from_yaml_config(
                    name=name, class_name=class_name, config=config_with_substitutions
                )
            elif class_name in self.TEST_YAML_CONFIG_SUPPORTED_DATASOURCE_TYPES:
                (
                    instantiated_class,
                    usage_stats_event_payload,
                ) = self._test_instantiation_of_datasource_from_yaml_config(
                    name=name,
                    class_name=class_name,
                    config=config,  # Uses original config as substitutions are done downstream
                )
            elif class_name in self.TEST_YAML_CONFIG_SUPPORTED_CHECKPOINT_TYPES:
                (
                    instantiated_class,
                    usage_stats_event_payload,
                ) = self._test_instantiation_of_checkpoint_from_yaml_config(
                    name=name, class_name=class_name, config=config_with_substitutions
                )
            elif class_name in self.TEST_YAML_CONFIG_SUPPORTED_DATA_CONNECTOR_TYPES:
                (
                    instantiated_class,
                    usage_stats_event_payload,
                ) = self._test_instantiation_of_data_connector_from_yaml_config(
                    name=name,
                    class_name=class_name,
                    config=config_with_substitutions,
                    runtime_environment=runtime_environment,
                )
            elif class_name in self.TEST_YAML_CONFIG_SUPPORTED_PROFILER_TYPES:
                (
                    instantiated_class,
                    usage_stats_event_payload,
                ) = self._test_instantiation_of_profiler_from_yaml_config(
                    name=name, class_name=class_name, config=config_with_substitutions
                )
            else:
                (
                    instantiated_class,
                    usage_stats_event_payload,
                ) = self._test_instantiation_of_misc_class_from_yaml_config(
                    name=name,
                    config=config_with_substitutions,
                    runtime_environment=runtime_environment,
                    usage_stats_event_payload=usage_stats_event_payload,
                )

            send_usage_message_from_handler(
                event=usage_stats_event_name,
                handler=self.usage_statistics_handler,
                event_payload=usage_stats_event_payload,
                success=True,
            )
            if pretty_print:
                print(
                    f"\tSuccessfully instantiated {instantiated_class.__class__.__name__}\n"
                )

            report_object: dict = instantiated_class.self_check(
                pretty_print=pretty_print
            )

            if return_mode == "instantiated_class":
                return instantiated_class

            return report_object

        except Exception as e:
            if class_name is None:
                usage_stats_event_payload[
                    "diagnostic_info"
                ] = usage_stats_event_payload.get(
                    "diagnostic_info", []
                ) + [  # type: ignore[operator]
                    "__class_name_not_provided__"
                ]
            elif (
                usage_stats_event_payload.get("parent_class") is None
                and class_name in self.ALL_TEST_YAML_CONFIG_SUPPORTED_TYPES
            ):
                # add parent_class if it doesn't exist and class_name is one of our supported core GX types
                usage_stats_event_payload["parent_class"] = class_name
            send_usage_message_from_handler(
                event=usage_stats_event_name,
                handler=self.usage_statistics_handler,
                event_payload=usage_stats_event_payload,
                success=False,
            )
            if shorten_tracebacks:
                traceback.print_exc(limit=1)
            else:
                raise e

    def _test_yaml_config_prepare_config(
        self, yaml_config: str, usage_stats_event_name: str
    ) -> CommentedMap:
        config = self._load_config_string_as_commented_map(
            config_str=yaml_config,
            usage_stats_event_name=usage_stats_event_name,
        )
        return config

    def _test_yaml_config_prepare_substituted_config(
        self, yaml_config: str, runtime_environment: dict, usage_stats_event_name: str
    ) -> CommentedMap:
        """
        Performs variable substitution and conversion from YAML to CommentedMap.
        See `test_yaml_config` for more details.
        """
        config_str_with_substituted_variables = (
            self._prepare_config_string_with_substituted_variables(
                yaml_config=yaml_config,
                runtime_environment=runtime_environment,
                usage_stats_event_name=usage_stats_event_name,
            )
        )
        config = self._load_config_string_as_commented_map(
            config_str=config_str_with_substituted_variables,
            usage_stats_event_name=usage_stats_event_name,
        )
        return config

    def _prepare_config_string_with_substituted_variables(
        self, yaml_config: str, runtime_environment: dict, usage_stats_event_name: str
    ) -> str:
        try:
            config_provider = self._data_context.config_provider
            config_values = config_provider.get_values()

            # While normally we'd just call `self.config_provider.substitute_config()`,
            # we need to account for `runtime_environment` values that may have been passed.
            config_values.update(runtime_environment)

            return config_provider.substitute_config(
                config=yaml_config, config_values=config_values
            )
        except Exception as e:
            usage_stats_event_payload: dict = {
                "diagnostic_info": ["__substitution_error__"],
            }
            send_usage_message_from_handler(
                event=usage_stats_event_name,
                handler=self.usage_statistics_handler,
                event_payload=usage_stats_event_payload,
                success=False,
            )
            raise e

    def _load_config_string_as_commented_map(
        self, config_str: str, usage_stats_event_name: str
    ) -> CommentedMap:
        try:
            substituted_config: CommentedMap = yaml.load(config_str)
            return substituted_config

        except Exception as e:
            usage_stats_event_payload = {
                "diagnostic_info": ["__yaml_parse_error__"],
            }
            send_usage_message_from_handler(
                event=usage_stats_event_name,
                handler=self.usage_statistics_handler,
                event_payload=usage_stats_event_payload,
                success=False,
            )
            raise e

    def _test_instantiation_of_store_from_yaml_config(
        self, name: Optional[str], class_name: str, config: CommentedMap
    ) -> Tuple[Store, dict]:
        """
        Helper to create store instance and update usage stats payload.
        See `test_yaml_config` for more details.
        """
        print(f"\tInstantiating as a Store, since class_name is {class_name}")
        store_name: str = name or config.get("name") or "my_temp_store"
        instantiated_class = self._data_context._build_store_from_config(
            store_name=store_name,
            store_config=config,
        )
        store_name = instantiated_class.store_name or store_name
        self._data_context.config["stores"][store_name] = config

        anonymizer = Anonymizer(self._data_context.data_context_id)
        usage_stats_event_payload = anonymizer.anonymize(
            store_name=store_name, store_obj=instantiated_class  # type: ignore[arg-type]
        )
        return instantiated_class, usage_stats_event_payload

    def _test_instantiation_of_datasource_from_yaml_config(
        self, name: Optional[str], class_name: str, config: CommentedMap
    ) -> Tuple[Datasource, dict]:
        """
        Helper to create datasource instance and update usage stats payload.
        See `test_yaml_config` for more details.
        """
        print(f"\tInstantiating as a Datasource, since class_name is {class_name}")
        datasource_name: str = name or config.get("name") or "my_temp_datasource"
        datasource_config = datasourceConfigSchema.load(config)
        datasource_config.name = datasource_name
        instantiated_class = cast(
            Datasource,
            self._data_context._instantiate_datasource_from_config_and_update_project_config(
                config=datasource_config,
                initialize=True,
                save_changes=False,
            ),
        )

        anonymizer = Anonymizer(self._data_context.data_context_id)

        if class_name == "SimpleSqlalchemyDatasource":
            # Use the raw config here, defaults will be added in the anonymizer
            usage_stats_event_payload = anonymizer.anonymize(
                obj=instantiated_class, name=datasource_name, config=config  # type: ignore[arg-type]
            )
        else:
            # Roundtrip through schema validation to remove any illegal fields add/or restore any missing fields.
            datasource_config = datasourceConfigSchema.load(instantiated_class.config)
            full_datasource_config = datasourceConfigSchema.dump(datasource_config)
            usage_stats_event_payload = anonymizer.anonymize(
                obj=instantiated_class,
                name=datasource_name,  # type: ignore[arg-type]
                config=full_datasource_config,
            )
        return instantiated_class, usage_stats_event_payload

    def _test_instantiation_of_checkpoint_from_yaml_config(
        self, name: Optional[str], class_name: str, config: CommentedMap
    ) -> Tuple[Checkpoint, dict]:
        """
        Helper to create checkpoint instance and update usage stats payload.
        See `test_yaml_config` for more details.
        """
        print(f"\tInstantiating as a {class_name}, since class_name is {class_name}")

        checkpoint_name: str = name or config.get("name") or "my_temp_checkpoint"

        checkpoint_config: Union[CheckpointConfig, dict]

        checkpoint_config = CheckpointConfig.from_commented_map(commented_map=config)
        checkpoint_config_dict: dict[str, JSONValues] = checkpoint_config.to_json_dict()
        checkpoint_config_dict.update({"name": checkpoint_name})

        checkpoint_class_args: dict = filter_properties_dict(  # type: ignore[assignment]
            properties=checkpoint_config_dict,
            delete_fields={"class_name", "module_name"},
            clean_falsy=True,
        )

        if class_name == "Checkpoint":
            instantiated_class = Checkpoint(
                data_context=self._data_context, **checkpoint_class_args
            )
        elif class_name == "SimpleCheckpoint":
            instantiated_class = SimpleCheckpoint(
                data_context=self._data_context, **checkpoint_class_args
            )
        else:
            raise ValueError(f'Unknown Checkpoint class_name: "{class_name}".')

        anonymizer = Anonymizer(self._data_context.data_context_id)

        usage_stats_event_payload = anonymizer.anonymize(
            obj=instantiated_class, name=checkpoint_name, config=checkpoint_config_dict  # type: ignore[arg-type]
        )

        return instantiated_class, usage_stats_event_payload

    def _test_instantiation_of_data_connector_from_yaml_config(
        self,
        name: Optional[str],
        class_name: str,
        config: CommentedMap,
        runtime_environment: dict,
    ) -> Tuple[DataConnector, dict]:
        """
        Helper to create data connector instance and update usage stats payload.
        See `test_yaml_config` for more details.
        """
        print(f"\tInstantiating as a DataConnector, since class_name is {class_name}")
        data_connector_name: str = (
            name or config.get("name") or "my_temp_data_connector"
        )
        instantiated_class = instantiate_class_from_config(
            config=config,
            runtime_environment={
                **runtime_environment,
                **{
                    "root_directory": self._data_context.root_directory,
                },
            },
            config_defaults={},
        )

        anonymizer = Anonymizer(self._data_context.data_context_id)

        usage_stats_event_payload = anonymizer.anonymize(
            obj=instantiated_class, name=data_connector_name, config=config  # type: ignore[arg-type]
        )
        return instantiated_class, usage_stats_event_payload

    def _test_instantiation_of_profiler_from_yaml_config(
        self, name: Optional[str], class_name: str, config: CommentedMap
    ) -> Tuple[RuleBasedProfiler, dict]:
        """
        Helper to create profiler instance and update usage stats payload.
        See `test_yaml_config` for more details.
        """
        print(f"\tInstantiating as a {class_name}, since class_name is {class_name}")

        profiler_name: str = name or config.get("name") or "my_temp_profiler"

        profiler_config: Union[
            RuleBasedProfilerConfig, dict
        ] = RuleBasedProfilerConfig.from_commented_map(commented_map=config)
        profiler_config = profiler_config.to_json_dict()  # type: ignore[union-attr]
        profiler_config.update({"name": profiler_name})

        instantiated_class = instantiate_class_from_config(
            config=profiler_config,
            runtime_environment={"data_context": self._data_context},
            config_defaults={
                "module_name": "great_expectations.rule_based_profiler",
                "class_name": "RuleBasedProfiler",
            },
        )

        anonymizer = Anonymizer(self._data_context.data_context_id)

        usage_stats_event_payload: dict = anonymizer.anonymize(
            obj=instantiated_class, name=profiler_name, config=profiler_config  # type: ignore[arg-type]
        )

        return instantiated_class, usage_stats_event_payload

    def _test_instantiation_of_misc_class_from_yaml_config(
        self,
        name: Optional[str],
        config: CommentedMap,
        runtime_environment: dict,
        usage_stats_event_payload: dict,
    ) -> Tuple[Any, dict]:
        """
        Catch-all to cover all classes not covered in other `_test_instantiation` methods.
        Attempts to match config to the relevant class/parent and update usage stats payload.
        See `test_yaml_config` for more details.
        """
        print(
            "\tNo matching class found. Attempting to instantiate class from the raw config..."
        )
        instantiated_class = instantiate_class_from_config(
            config=config,
            runtime_environment={
                **runtime_environment,
                **{
                    "root_directory": self._data_context.root_directory,
                },
            },
            config_defaults={},
        )

        # If a subclass of a supported type, find the parent class and anonymize
        anonymizer = Anonymizer(self._data_context.data_context_id)

        parent_class_from_object = anonymizer.get_parent_class(
            object_=instantiated_class
        )
        parent_class_from_config = anonymizer.get_parent_class(object_config=config)

        if parent_class_from_object is not None and parent_class_from_object.endswith(
            "Store"
        ):
            store_name: str = name or config.get("name") or "my_temp_store"
            store_name = instantiated_class.store_name or store_name
            usage_stats_event_payload = anonymizer.anonymize(
                store_name=store_name, store_obj=instantiated_class  # type: ignore[arg-type]
            )
        elif parent_class_from_config is not None and parent_class_from_config.endswith(
            "Datasource"
        ):
            datasource_name: str = name or config.get("name") or "my_temp_datasource"
            if DatasourceAnonymizer.get_parent_class_v3_api(config=config):
                # Roundtrip through schema validation to remove any illegal fields add/or restore any missing fields.
                datasource_config = datasourceConfigSchema.load(
                    instantiated_class.config
                )
                full_datasource_config = datasourceConfigSchema.dump(datasource_config)
            else:
                # for v2 api
                full_datasource_config = config
            if parent_class_from_config == "SimpleSqlalchemyDatasource":
                # Use the raw config here, defaults will be added in the anonymizer
                usage_stats_event_payload = anonymizer.anonymize(
                    obj=instantiated_class, name=datasource_name, config=config  # type: ignore[arg-type]
                )
            else:
                usage_stats_event_payload = anonymizer.anonymize(
                    obj=instantiated_class,
                    name=datasource_name,  # type: ignore[arg-type]
                    config=full_datasource_config,
                )

        elif parent_class_from_config is not None and parent_class_from_config.endswith(
            "Checkpoint"
        ):
            checkpoint_name: str = name or config.get("name") or "my_temp_checkpoint"
            # Roundtrip through schema validation to remove any illegal fields add/or restore any missing fields.
            checkpoint_config: Union[CheckpointConfig, dict]
            checkpoint_config = CheckpointConfig.from_commented_map(
                commented_map=config
            )
            checkpoint_config_dict: dict[
                str, JSONValues
            ] = checkpoint_config.to_json_dict()
            checkpoint_config_dict.update({"name": checkpoint_name})
            usage_stats_event_payload = anonymizer.anonymize(
                obj=checkpoint_config_dict, name=checkpoint_name, config=checkpoint_config  # type: ignore[arg-type]
            )

        elif parent_class_from_config is not None and parent_class_from_config.endswith(
            "DataConnector"
        ):
            data_connector_name: str = (
                name or config.get("name") or "my_temp_data_connector"
            )
            usage_stats_event_payload = anonymizer.anonymize(
                obj=instantiated_class, name=data_connector_name, config=config  # type: ignore[arg-type]
            )

        else:
            # If class_name is not a supported type or subclass of a supported type,
            # mark it as custom with no additional information since we can't anonymize
            usage_stats_event_payload[
                "diagnostic_info"
            ] = usage_stats_event_payload.get("diagnostic_info", []) + [
                "__custom_subclass_not_core_ge__"
            ]

        return instantiated_class, usage_stats_event_payload



from great_expectations.util import filter_properties_dict


def test_data_asset_citations(pandas_dataset):
    citation_date = "2020-02-27T12:34:56.123456Z"
    pandas_dataset.add_citation("test citation", citation_date=citation_date)
    suite = pandas_dataset.get_expectation_suite()
    assert filter_properties_dict(
        properties=suite.meta["citations"][0],
        delete_fields={
            "interactive",
        },
        clean_falsy=True,
    ) == filter_properties_dict(
        properties={
            "comment": "test citation",
            "batch_kwargs": pandas_dataset.batch_kwargs,
            "batch_parameters": pandas_dataset.batch_parameters,
            "batch_markers": pandas_dataset.batch_markers,
            "citation_date": citation_date,
        },
        clean_falsy=True,
    )



# <snippet name="tests/integration/docusaurus/connecting_to_your_data/database/sqlite_yaml_example.py imports">
from ruamel import yaml

import great_expectations as gx
from great_expectations.core.batch import BatchRequest, RuntimeBatchRequest

# </snippet>

CONNECTION_STRING = "sqlite:///data/yellow_tripdata.db"

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/database/sqlite_yaml_example.py get_context">
context = gx.get_context()
# </snippet>

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/database/sqlite_yaml_example.py datasource_yaml">
datasource_yaml = f"""
name: my_sqlite_datasource
class_name: Datasource
execution_engine:
  class_name: SqlAlchemyExecutionEngine
  connection_string: sqlite://<PATH_TO_DB_FILE>
data_connectors:
   default_runtime_data_connector_name:
       class_name: RuntimeDataConnector
       batch_identifiers:
           - default_identifier_name
   default_inferred_data_connector_name:
       class_name: InferredAssetSqlDataConnector
       include_schema_name: true
"""
# </snippet>

# Please note this override is only to provide good UX for docs and tests.
# In normal usage you'd set your path directly in the yaml above.
datasource_yaml = datasource_yaml.replace(
    "sqlite://<PATH_TO_DB_FILE>",
    CONNECTION_STRING,
)
# <snippet name="tests/integration/docusaurus/connecting_to_your_data/database/sqlite_yaml_example.py test_yaml_config">
context.test_yaml_config(datasource_yaml)
# </snippet>

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/database/sqlite_yaml_example.py add_datasource">
context.add_datasource(**yaml.load(datasource_yaml))
# </snippet>

# Here is a RuntimeBatchRequest using a query
# <snippet name="tests/integration/docusaurus/connecting_to_your_data/database/sqlite_yaml_example.py query batch_request">
batch_request = RuntimeBatchRequest(
    datasource_name="my_sqlite_datasource",
    data_connector_name="default_runtime_data_connector_name",
    data_asset_name="default_name",  # this can be anything that identifies this data
    runtime_parameters={
        "query": "SELECT * from main.yellow_tripdata_sample_2019_01 LIMIT 10"
    },
    batch_identifiers={"default_identifier_name": "default_identifier"},
)
context.add_or_update_expectation_suite(expectation_suite_name="test_suite")
validator = context.get_validator(
    batch_request=batch_request, expectation_suite_name="test_suite"
)
print(validator.head())
# </snippet>

# NOTE: The following code is only for testing and can be ignored by users.
assert isinstance(validator, gx.validator.validator.Validator)

# Here is a BatchRequest naming a table
batch_request = BatchRequest(
    datasource_name="my_sqlite_datasource",
    data_connector_name="default_inferred_data_connector_name",
    data_asset_name="main.yellow_tripdata_sample_2019_01",  # this is the name of the table you want to retrieve
)
context.add_or_update_expectation_suite(expectation_suite_name="test_suite")
validator = context.get_validator(
    batch_request=batch_request, expectation_suite_name="test_suite"
)
print(validator.head())

# NOTE: The following code is only for testing and can be ignored by users.
assert isinstance(validator, gx.validator.validator.Validator)
assert [ds["name"] for ds in context.list_datasources()] == ["my_sqlite_datasource"]
assert "main.yellow_tripdata_sample_2019_01" in set(
    context.get_available_data_asset_names()["my_sqlite_datasource"][
        "default_inferred_data_connector_name"
    ]
)



from typing import Any, Optional

from great_expectations.core.usage_statistics.anonymizers.base import BaseAnonymizer
from great_expectations.datasource import (
    BaseDatasource,
    Datasource,
    LegacyDatasource,
    PandasDatasource,
    SimpleSqlalchemyDatasource,
    SparkDFDatasource,
    SqlAlchemyDatasource,
)


class DatasourceAnonymizer(BaseAnonymizer):
    # ordered bottom up in terms of inheritance order
    _legacy_ge_classes = [
        PandasDatasource,
        SqlAlchemyDatasource,
        SparkDFDatasource,
        LegacyDatasource,
    ]

    # ordered bottom up in terms of inheritance order
    _ge_classes = [
        SimpleSqlalchemyDatasource,
        Datasource,
        BaseDatasource,
    ]

    def __init__(
        self,
        aggregate_anonymizer: "Anonymizer",  # noqa: F821
        salt: Optional[str] = None,
    ) -> None:
        super().__init__(salt=salt)

        self._aggregate_anonymizer = aggregate_anonymizer

    def anonymize(self, obj: Optional[object] = None, *args, **kwargs) -> Any:
        if obj is not None and isinstance(obj, SimpleSqlalchemyDatasource):
            return self._anonymize_simple_sqlalchemy_datasource(*args, **kwargs)
        return self._anonymize_datasource_info(*args, **kwargs)

    def _anonymize_datasource_info(self, name: str, config: dict) -> dict:
        anonymized_info_dict = {}
        anonymized_info_dict["anonymized_name"] = self._anonymize_string(name)

        # Legacy Datasources (<= v0.12 v2 BatchKwargs API)
        if self.get_parent_class_v2_api(config=config) is not None:
            self._anonymize_object_info(
                anonymized_info_dict=anonymized_info_dict,
                object_config=config,
            )
        # Datasources (>= v0.13 v3 BatchRequest API), and custom v2 BatchKwargs API
        elif self.get_parent_class_v3_api(config=config) is not None:
            self._anonymize_object_info(
                anonymized_info_dict=anonymized_info_dict,
                object_config=config,
            )
            execution_engine_config = config.get("execution_engine")
            anonymized_info_dict[
                "anonymized_execution_engine"
            ] = self._anonymize_execution_engine_info(
                name=execution_engine_config.get("name", ""),
                config=execution_engine_config,
            )
            data_connector_configs = config.get("data_connectors")
            anonymized_info_dict["anonymized_data_connectors"] = [
                self._aggregate_anonymizer.anonymize(
                    name=data_connector_name, config=data_connector_config
                )
                for data_connector_name, data_connector_config in data_connector_configs.items()
            ]

        return anonymized_info_dict

    def _anonymize_simple_sqlalchemy_datasource(self, name: str, config: dict) -> dict:
        """
        SimpleSqlalchemyDatasource requires a separate anonymization scheme.
        """
        anonymized_info_dict = {}
        anonymized_info_dict["anonymized_name"] = self._anonymize_string(name)
        if config.get("module_name") is None:
            config["module_name"] = "great_expectations.datasource"
        self._anonymize_object_info(
            anonymized_info_dict=anonymized_info_dict,
            object_config=config,
        )

        # Only and directly provide parent_class of execution engine
        anonymized_info_dict["anonymized_execution_engine"] = {
            "parent_class": "SqlAlchemyExecutionEngine"
        }

        # Use the `introspection` and `tables` keys to find data_connectors in SimpleSqlalchemyDatasources
        introspection_data_connector_configs = config.get("introspection")
        tables_data_connector_configs = config.get("tables")

        introspection_data_connector_anonymized_configs = []
        if introspection_data_connector_configs is not None:
            for (
                data_connector_name,
                data_connector_config,
            ) in introspection_data_connector_configs.items():
                if data_connector_config.get("class_name") is None:
                    data_connector_config[
                        "class_name"
                    ] = "InferredAssetSqlDataConnector"
                if data_connector_config.get("module_name") is None:
                    data_connector_config[
                        "module_name"
                    ] = "great_expectations.datasource.data_connector"
                introspection_data_connector_anonymized_configs.append(
                    self._aggregate_anonymizer.anonymize(
                        name=data_connector_name, config=data_connector_config
                    )
                )

        tables_data_connector_anonymized_configs = []
        if tables_data_connector_configs is not None:
            for (
                data_connector_name,
                data_connector_config,
            ) in tables_data_connector_configs.items():
                if data_connector_config.get("class_name") is None:
                    data_connector_config[
                        "class_name"
                    ] = "ConfiguredAssetSqlDataConnector"
                if data_connector_config.get("module_name") is None:
                    data_connector_config[
                        "module_name"
                    ] = "great_expectations.datasource.data_connector"
                tables_data_connector_anonymized_configs.append(
                    self._aggregate_anonymizer.anonymize(
                        name=data_connector_name, config=data_connector_config
                    )
                )

        anonymized_info_dict["anonymized_data_connectors"] = (
            introspection_data_connector_anonymized_configs
            + tables_data_connector_anonymized_configs
        )

        return anonymized_info_dict

    def _anonymize_execution_engine_info(self, name: str, config: dict) -> dict:
        anonymized_info_dict = {}
        anonymized_info_dict["anonymized_name"] = self._anonymize_string(name)

        from great_expectations.data_context.types.base import (
            ExecutionEngineConfig,
            executionEngineConfigSchema,
        )

        # Roundtrip through schema validation to remove any illegal fields add/or restore any missing fields.
        execution_engine_config: ExecutionEngineConfig = (
            executionEngineConfigSchema.load(config)
        )
        execution_engine_config_dict: dict = executionEngineConfigSchema.dump(
            execution_engine_config
        )

        self._anonymize_object_info(
            anonymized_info_dict=anonymized_info_dict,
            object_config=execution_engine_config_dict,
        )

        return anonymized_info_dict

    def can_handle(self, obj: Optional[object] = None, **kwargs) -> bool:
        return obj is not None and isinstance(obj, BaseDatasource)

    @staticmethod
    def get_parent_class(config: dict) -> Optional[str]:
        return BaseAnonymizer.get_parent_class(
            classes_to_check=DatasourceAnonymizer._ge_classes
            + DatasourceAnonymizer._legacy_ge_classes,
            object_config=config,
        )

    @staticmethod
    def get_parent_class_v2_api(config: dict) -> Optional[str]:
        return BaseAnonymizer.get_parent_class(
            classes_to_check=DatasourceAnonymizer._legacy_ge_classes,
            object_config=config,
        )

    @staticmethod
    def get_parent_class_v3_api(config: dict) -> Optional[str]:
        return BaseAnonymizer.get_parent_class(
            classes_to_check=DatasourceAnonymizer._ge_classes,
            object_config=config,
        )
